import vergeml.glossary as glossary
from vergeml.utils import format_info_text, did_you_mean, VergeMLError
from vergeml.command import Command, command, CommandPlugin
from vergeml.option import option
from vergeml.option import Option
from vergeml.operation import Operation
from vergeml.io import Source
from vergeml.model import Model
from vergeml.commands.preprocess import PreprocessCommand
from vergeml.random_robot import random_robot_name, ascii_robot
import webbrowser
import pydoc
import io
import textwrap
import random
import datetime
from typing import Optional

USAGE= """
Usage: ml [--version] [--file=<path>] [--model=<name>] [--samples-dir=<path>]
          [--val-split=<conf>] [--test-split=<conf>] [--cache-dir=<path>]
          [--trainings-dir=<path>] [--project-dir=<path>] [--random-seed=<value>]
          [--cache=<conf>] [--device=<id>] [--device-memory=<conf>]
          [@AI ...] <command> [<args> ...]
""".strip()

HELP_OPTIONS = [
    ('-v, --version',    'Print the version and exit.'),
    ('-f, --file',       'The path to the project file [default vergeml.yaml].'),
    ('-m, --model',      'The model to train.'),
    ('--samples-dir',    'The name of the directory where samples are stored [default: samples].'),
    ('--val-split',      'Validation split configuration [default: 10%].' ),
    ('--test-split',     'Test split configuration [default: 10%].'),
    ('--cache-dir',      'The cache directory [default: .cache].'),
    ('--trainings-dir',  'The name of the directory the training data is written to [default: trainings]'),
    ('--project-dir',    'The path of the project directory [default: the current directory].'),
    ('--random-seed',    'Seed of the random generator [default: 42].'),
    ('--cache',          'How to cache samples during training [default: mem]'),
    ('--device',         'Which device to use for training [default: auto].'),
    ('--device-memory',  'How much device memory to use on a GPU. [default: auto]'),
]


_GENERAL_HELP = [
    ('tutorial', 'Open the quickstart tutorial.'),
    ('options', 'List and describe VergeML options.'),
    ('glossary', 'Show a description of terms and concepts.'),
    ('models', 'List available VergeML models.')
]

_CONFIGURATION_HELP = [
    ('data', 'Describe the data section in the config file.'),
    ('input', 'How to configure sample input.'),
    ('preprocess', 'How to set up sample preprocessing.'),
    ('cache', 'Show cache options.'),
    ('output', 'Describe the data output option.'),
    ('device', 'How to configure CPU and GPU usage.')
]

_TUTORIAL_URL = "https://vergeml.com/tutorial"

_CACHE_HELP = """
Caching can help to speed up the training process. For example:

  data:
    cache: mem-out

mem:         Cache sample data in memory (default).
mem-out:     Cache data in memory after preprocessing.
disk:        Cache data on disk in a format optimized for fast access.
disk-out:    Cache data on disk after preprocessing.
none:        Don't cache.

You can configure the cache from your project file or on the command line via the --cache option.
"""

_OUTPUT_HELP = """
Output is an advanced option that can be set via the data.output option in the project file.
"""

_PREPROCESS_HELP = """
To change and augment your samples data during training, you can define a preprocessing pipeline. For example:

  data:
    preprocess:
       - op: augment
         variants: 4
       - op: random-crop
         width: 128
         height: 128
"""

_DATA_HELP = """
The data section in your config file lets you set up how data is loaded and processed. For example:

  data:
      input:
        type: images
        input-patterns: **/*.jpg
      preprocess:
        - op: center-crop
          width: 128
          height: 128

It is comprised of four subsections:

input:         Controls how the data is loaded.
preprocess:    Transform and augment sample data.
output:        Set the final transformation before training.
cache:         Use cache to speed up the training process.

To learn more, see 'ml help <subsection>', e.g. 'ml help preprocess'.
"""


_DEVICE_HELP = """
The device section in your config file lets you configure the device to use when training and predicting. It supports the following settings:

id:        The id of the device to use. (See below.)
memory:    How much device memory to use as a percentage. i.e. 50%
grow-memory:    When true, allow growing memory usage as needed.

The value of device.id can be either 'auto' (the default), 'cpu', 'gpu' or 'gpu:number', where number is the number of the GPU you want to use.

As a shortcut, you can set the device id directly, e.g.

  device: gpu

Device ID and memory usage can be set on the command line via --device and --device-memory flags.
"""

@command('help', descr="Get general help or specific help for a command.", free_form=True)
@option('<topic>', type=Optional[str], descr="The topic to get help on.")
@option('@AI', type='Optional[@]')
@option('all', type='Optional[bool]', short="a", flag=True, descr="Show all help topics.")
class HelpCommand(CommandPlugin):

    def __call__(self, args, env):
        _AI, topic = args
        topic = " ".join(topic)
        if topic == "tutorial":
            if webbrowser.open(_TUTORIAL_URL, new=2):
                print("Opened {} in a new browser tab.".format(_TUTORIAL_URL))
            else:
                print("Could not access the web browser. Please use this URL to access the tutorial:")
                print(_TUTORIAL_URL)
        else:
            print(self.get_help(env, topic))

    def get_help(self, env=None, topic="", short=False):

        if topic:
            model_commands = {}
            if env and env.model_plugin:
                for fn in Command.find_functions(env.model_plugin):
                    cmd = Command.discover(fn)
                    model_commands[cmd.name] = cmd

            if topic == "-a":
                return self.format_topics(env)
            # show VergeML options
            elif topic == "options":
                return self.format_options()

            # display the glossary
            elif topic == "glossary":
                return self.format_glossary()

            # show available models
            elif topic == "models":
                return _with_header(self.format_models(),
                                    help="models", topic=topic)

            # explain the data.input section
            elif topic == "input":
                return _with_header(self.format_input_list(),
                                    help="configuration", topic=topic)

            # explain the data.cache section
            elif topic == "cache":
                return _with_header(format_info_text(_CACHE_HELP),
                                    help="configuration", topic=topic)

            # explain the data.output section
            elif topic == "output":
                return _with_header(format_info_text(_OUTPUT_HELP),
                                    help="configuration", topic=topic)

            # explain preprocessing
            elif topic in ("preprocess", "preprocessing"):
                return _with_header(self.format_preprocessing_list(),
                                    help="configuration", topic=topic)

            # explain the data section
            elif topic == "data":
                return _with_header(format_info_text(_DATA_HELP.strip()),
                                    help="configuration", topic=topic)

            # explain the device section
            elif topic == "device":
                return _with_header(format_info_text(_DEVICE_HELP.strip()),
                                    help="configuration", topic=topic)

            # show a random robot
            elif topic == "random robot":
                robot = ascii_robot(datetime.datetime.now(),
                                    random_robot_name(datetime.datetime.now()))
                return f"\n{robot}\n"

            elif ":" in topic and topic.split(":", 1)[0] in self.plugins.keys('vergeml.cmd'):
                command, subcommand = topic.split(":")
                cmd = Command.discover(self.plugins.get('vergeml.cmd', command))
                subcommand_option = next(filter(lambda o: bool(o.subcommand), cmd.options), None)
                if not subcommand_option:
                    raise VergeMLError(f"{command} takes no subcommand", help_topic=command)

                plugin = self.plugins.get(subcommand_option.subcommand, subcommand)
                if not plugin:
                    raise VergeMLError(f"Invalid {subcommand_option.name}", help_topic=command)

                cmd = Command.discover(plugin)
                return cmd.usage(short, parent_command=command)

            # display model command help
            elif topic in model_commands:
                return model_commands[topic].usage(short)

            # display command help
            elif topic in self.plugins.keys('vergeml.cmd'):
                cmd = Command.discover(self.plugins.get('vergeml.cmd', topic))
                return cmd.usage(short)

            elif topic in self.plugins.keys('vergeml.operation'):
                return _with_header(self.format_source_or_operation(topic, 'vergeml.operation', Operation),
                                    help="preprocessing operation", topic=topic)

            elif topic in self.plugins.keys('vergeml.io'):
                return _with_header(self.format_source_or_operation(topic, 'vergeml.io', Source),
                                    help="data source", topic=topic)

            elif topic in self.plugins.keys('vergeml.model'):
                return _with_header(self.format_model(topic),
                                    help="models", topic=topic)

            # show a glossary entry
            elif glossary.long_descr(topic):
                topic = glossary.SYNONYMS.get(topic, topic)
                return _with_header(format_info_text(glossary.long_descr(topic)),
                                    help="glossary", topic=topic)

            # show base options help
            elif topic in dict(HELP_OPTIONS):
                return _with_header(format_info_text(dict(HELP_OPTIONS).get(topic)),
                                    help="base options", topic=topic)

            else:
                candidates = set()
                candidates.update(map(lambda h: h[0], _GENERAL_HELP))
                candidates.update(self.plugins.keys("vergeml.cmd"))
                candidates.update(map(lambda h: h[0], _CONFIGURATION_HELP))
                candidates.update(self.plugins.keys("vergeml.io"))
                candidates.update(self.plugins.keys("vergeml.operation"))
                candidates.update(self.plugins.keys("vergeml.model"))
                if env and env.model_plugin:
                    for fn in Command.find_functions(env.model_plugin):
                        cmd = Command.discover(fn)
                        candidates.add(cmd.name)
                candidates.update(glossary.LONG_DESCR.keys())
                candidates.update(glossary.SYNONYMS.keys())

                suggestion = did_you_mean(list(candidates), topic)
                if suggestion:
                    return f"No help found for topic '{topic}'. " + suggestion
                else:
                    return f"No help found for topic '{topic}'."

        else:
            return self.format_general_help(env, short=short)

    def format_topics(self, env):
        buffer = io.StringIO()

        print("General Help:", file=buffer)
        for topic, descr in _GENERAL_HELP:
            print("  {:<16} {}".format(topic, descr), file=buffer)
        print("", file=buffer)

        print("Commands:", file=buffer)
        for cmd_name in self.plugins.keys('vergeml.cmd'):
            descr = Command.discover(self.plugins.get('vergeml.cmd', cmd_name)).descr
            print("  {:<16} {}".format(cmd_name, descr), file=buffer)
        print("", file=buffer)

        if env and env.model_plugin:
            print("Model Commands:", file=buffer)
            for fn in Command.find_functions(env.model_plugin):
                cmd = Command.discover(fn)
                print("  {:<16} {}".format(cmd.name, cmd.descr), file=buffer)
            print("", file=buffer)

        print("Configuration:", file=buffer)
        for topic, descr in _CONFIGURATION_HELP:
            print("  {:<16} {}".format(topic, descr), file=buffer)
        print("", file=buffer)

        inputs = []
        for k in self.plugins.keys('vergeml.io'):
            plugin = self.plugins.get('vergeml.io', k)
            source = Source.discover(plugin)
            inputs.append((k, source.descr))

        print("Data Input:", file=buffer)
        print(_get_table(inputs, IND=2, colon=False), file=buffer)
        print("", file=buffer)

        ops = {}
        for k in self.plugins.keys('vergeml.operation'):
            plugin = self.plugins.get('vergeml.operation', k)
            op = Operation.discover(plugin)
            topic = op.topic or "general"
            descr = op.descr
            ops.setdefault(topic, [])
            ops[topic].append((k, descr))

        for k, v in sorted(ops.items()):
            topic = k.capitalize()
            print(f"{topic} Operations:", file=buffer)
            print(format_info_text(_get_table(v), indent=2), file=buffer)
            print("", file=buffer)

        models = []
        for name in self.plugins.keys("vergeml.model"):
            plugin = self.plugins.get('vergeml.model', name)
            model = Model.discover(plugin)
            models.append((name, model.descr))

        if models:
            print(_get_table(models), file=buffer)

        print ("Glossary:", file=buffer)
        items = ", ".join(glossary.LONG_DESCR.keys())
        print(format_info_text(items, indent=2), file=buffer)

        return buffer.getvalue().strip()

    def format_glossary(self):
        buffer = io.StringIO()
        for k, v in sorted(glossary.LONG_DESCR.items()):
            print(f"{k}:", file=buffer)
            print(format_info_text(v, indent=2), file=buffer)
            print("", file=buffer)
        return buffer.getvalue().strip()

    def format_models(self):
        buffer = io.StringIO()

        model_names = self.plugins.keys("vergeml.model")
        if not model_names:
            print("No models installed.", file=buffer)
            # LATER
            # print("No models installed. See 'ml help install'", file=buffer)
            return buffer.getvalue().strip()
        models = []
        for name in model_names:
            plugin = self.plugins.get('vergeml.model', name)
            model = Model.discover(plugin)
            models.append((name, model.descr))

        print(_get_table(models), file=buffer)

        n_models = len(models)
        label = 'model' if n_models == 1 else 'models'
        print("", file=buffer)
        print(f"{n_models} {label} installed.", file=buffer)
        # LATER
        #print(f"{n_models} {label} installed. To install more, see 'ml help install'", file=buffer)
        return buffer.getvalue().strip()

    def format_general_help(self, env=None, short=False):
        buffer = io.StringIO()
        print(USAGE, file=buffer)
        print("", file=buffer)

        terms = set(glossary.LONG_DESCR.keys())
        terms.update(self.plugins.keys('vergeml.cmd'))
        rng = random.Random()
        rng.seed(datetime.datetime.now())
        random_term = rng.choice(list(terms))

        if env and env.model_plugin:
            print(f"Current Model: {env.get('model')}", file=buffer)
            print("", file=buffer)

        print("General Help:", file=buffer)
        for topic, descr in _GENERAL_HELP:
            print("  {:<16} {}".format(topic, descr), file=buffer)
        print("", file=buffer)

        print("Commands:", file=buffer)
        for cmd_name in self.plugins.keys('vergeml.cmd'):
            descr = Command.discover(self.plugins.get('vergeml.cmd', cmd_name)).descr
            print("  {:<16} {}".format(cmd_name, descr), file=buffer)
        print("", file=buffer)

        if env and env.model_plugin:
            print("Model Commands:", file=buffer)
            for fn in Command.find_functions(env.model_plugin):
                cmd = Command.discover(fn)
                print("  {:<16} {}".format(cmd.name, cmd.descr), file=buffer)
            print("", file=buffer)

        if not short:
            print("See 'ml help <command>' or 'ml help <topic>' to read about a specific subcommand or topic.", file=buffer)
            print(f"For example, try 'ml help {random_term}'", file=buffer)

        return buffer.getvalue().strip()

    def format_options(self):
        buffer = io.StringIO()
        print(USAGE, file=buffer)
        print("", file=buffer)
        for k, v in HELP_OPTIONS:
            print("{:<18} {}".format(k, v), file=buffer)
        return buffer.getvalue().strip()

    def format_input_list(self):
        buffer = io.StringIO()
        print("Controls how your data is loaded. Possible values for data.input are:", file=buffer)
        print("", file=buffer)
        inputs = []
        for k in self.plugins.keys('vergeml.io'):
            plugin = self.plugins.get('vergeml.io', k)
            source = Source.discover(plugin)
            inputs.append((k, source.descr))

        print(_get_table(inputs), file=buffer)
        print("", file=buffer)
        print("For more help on a particular input, see 'ml help <input-name>.'", file=buffer)
        return buffer.getvalue().strip()

    def format_preprocessing_list(self):
        buffer = io.StringIO()

        cmd = Command.discover(PreprocessCommand)
        print(cmd.usage(), file=buffer)
        print("", file=buffer)
        print("Configuration:", file=buffer)

        print(format_info_text(_PREPROCESS_HELP.strip(), indent=2), file=buffer)
        print("", file=buffer)

        ops = {}
        for k in self.plugins.keys('vergeml.operation'):
            plugin = self.plugins.get('vergeml.operation', k)
            op = Operation.discover(plugin)
            topic = op.topic or "general"
            descr = op.descr
            ops.setdefault(topic, [])
            ops[topic].append((k, descr))

        for k, v in sorted(ops.items()):
            topic = k.capitalize()
            print(f"{topic} Operations:", file=buffer)
            print(format_info_text(_get_table(v), indent=2), file=buffer)
            print("", file=buffer)

        return buffer.getvalue().strip()

    def format_source_or_operation(self, name, group, what):
        op = what.discover(self.plugins.get(group, name))
        buffer = io.StringIO()
        descr = op.long_descr or op.descr
        if descr:
            print(descr, file=buffer)
        else:
            print(name, file=buffer)
        print("", file=buffer)
        options = op.options
        if options:
            print("Options:", file=buffer)
            for opt in options:
                print("", file=buffer)
                tp_descr = opt.human_type()
                if tp_descr:
                    print(f"{opt.name}: {tp_descr}", file=buffer)
                else:
                    print(f"{opt.name}", file=buffer)
                if opt.descr:
                    print(format_info_text(opt.descr, indent=2), file=buffer)
        return buffer.getvalue().strip()

    def format_model(self, name):
        plugin = self.plugins.get("vergeml.model", name)
        model = Model.discover(plugin)
        buffer = io.StringIO()
        if model.descr:
            print(format_info_text(model.descr), file=buffer)

        if model.long_descr:
            if model.descr:
                print("", file=buffer)
            print(format_info_text(model.long_descr), file=buffer)
        return buffer.getvalue()

def _get_table(table, SPACE=4, IND=0, colon=True):
    buffer = io.StringIO()
    max_name = max(map(lambda o: len(o[0]), table))
    for name, descr in sorted(table):
        if descr:
            nspace = (max_name + SPACE) - len(name)
            space = str(nspace * ' ')
            col = ":" if colon else ""
            ind = str(" "*IND)
            print(f"{ind}{name}{col}{space}{descr}", file=buffer)
        else:
            print(name, file=buffer)
    return buffer.getvalue().rstrip()

def _with_header(content, help, topic):
    buffer = io.StringIO()
    print(f"HELP:  *{help}*", file=buffer)
    print(f"TOPIC: {topic}", file=buffer)
    print("", file=buffer)
    print(content, file=buffer)
    return buffer.getvalue().strip()



"""
Sample caching support.
"""

import struct
import pickle
import mmap
import io
import numpy as np
import lz4.frame

from vergeml import VergeMLError

class Cache:
    """Abstract base class for caches.
    """

    def write(self, data, meta):
        """Write data and metadata to the cache.
        """
        raise NotImplementedError

    def read(self, index, n_samples):
        """Read n_samples at index from the cache.
        """
        raise NotImplementedError

class MemoryCache(Cache):
    """Cache samples in memory.
    """

    def __init__(self):
        self.data = []

    def __len__(self):
        return len(self.data)

    def write(self, data, meta):
        self.data.append((data, meta))

    def read(self, index, n_samples):
        return self.data[index:index+n_samples]

class _CacheFileContent:

    def __init__(self):

        # An index of the positions of the stored data items.
        self.index = []

        # Sample metadata.
        self.meta = []

        # Info (Used to store data types)
        self.info = None

    def read(self, file, path):
        """Read the content index from file.
        """
        pos, = struct.unpack('<Q', file.read(8))
        if pos == 0:
            raise VergeMLError("Invalid cache file: {}".format(path))
        file.seek(pos)
        self.index, self.meta, self.info = pickle.load(file)

    def write(self, file):
        """Write the content index to file and update the header.
        """
        pos = file.tell()
        pickle.dump((self.index, self.meta, self.info), file)
        file.seek(0)

        # update the header with the position of the content index.
        file.write(struct.pack('<Q', pos))

class FileCache(Cache):
    """Cache raw bytes in a mmapped file.
    """

    def __init__(self, path, mode):
        assert mode in ("r", "w")

        self.path = path
        self.file = open(self.path, mode + "b")
        self.mmfile = None
        self.mode = mode
        self.cnt = _CacheFileContent()

        if mode == "r":
            # Read the last part of the file which contains the contents of the
            # cache.
            self.cnt.read(self.file, self.path)
            self.mmfile = mmap.mmap(self.file.fileno(), 0, access=mmap.ACCESS_READ)
        else:
            # The 8 bytes header contain the position of the content index.
            # We fill this header with zeroes and write the actual position
            # once all samples have been written to the cache
            self.file.write(struct.pack('<Q', 0))

    def __len__(self):
        return len(self.cnt.index)

    def write(self, data, meta):
        assert self.mode == "w"

        pos = self.file.tell()
        entry = (pos, pos + len(data))

        # write position and metadata of the data to the content index
        self.cnt.index.append(entry)
        self.cnt.meta.append(meta)
        self.file.write(data)

    def read(self, index, n_samples):
        assert self.mode == "r"

        c_ix = self.cnt.index

        # get the absolute start and end adresses of the whole chunk
        abs_start, _ = c_ix[index]
        _, abs_end = c_ix[index+n_samples-1]

        # read the bytes and wrap in memory view to avoid copying
        chunk = memoryview(self.mmfile[abs_start:abs_end])

        res = []

        for i in range(n_samples):
            start, end = c_ix[index+i]

            # convert addresses to be relative to the chunk we read
            start = start - abs_start
            end = end - abs_start

            data = chunk[start:end]
            res.append((data, self.cnt.meta[index+i]))
        return res

    def close(self):
        """Close the cache file.

        When the cache file is being written to, this method will write
        the content index at the end of the file.
        """
        if self.mode == "w":
            # Write the content index
            self.cnt.write(self.file)

        self.file.close()

# The three basic serialization methods:
# raw bytes, numpy format or python pickle.
_BYTES, _NUMPY, _PICKLE = range(3)

class SerializedFileCache(FileCache):
    """Cache serialized objects in a mmapped file.
    """

    def __init__(self, path, mode, compress=True):
        """Create an optionally compressed serialized cache.
        """
        super().__init__(path, mode)

        # we use info to store type information
        self.cnt.info = self.cnt.info or []
        self.compress = compress

    def _serialize_data(self, data):

        # Default to raw bytes
        type_ = _BYTES

        if isinstance(data, np.ndarray):
        # When the data is a numpy array, use the more compact native
        # numpy format.
            buf = io.BytesIO()
            np.save(buf, data)
            data = buf.getvalue()
            type_ = _NUMPY

        elif not isinstance(data, (bytearray, bytes)):
        # Everything else except byte data is serialized in pickle format.
            data = pickle.dumps(data)
            type_ = _PICKLE

        if self.compress:
        # Optional compression
            data = lz4.frame.compress(data)

        return type_, data

    def _deserialize(self, data, type_):

        if self.compress:
        # decompress the data if needed
            data = lz4.frame.decompress(data)

        if type_ == _NUMPY:
        # deserialize numpy arrays
            buf = io.BytesIO(data)
            data = np.load(buf)

        elif type_ == _PICKLE:
        # deserialize other python objects
            data = pickle.loads(data)

        else:
        # Otherwise we just return data as it is (bytes)
            pass

        return data

    def write(self, data, meta):

        if isinstance(data, tuple) and len(data) == 2:
            # write (x,y) pairs

            # serialize independent from each other
            type1, data1 = self._serialize_data(data[0])
            type2, data2 = self._serialize_data(data[1])

            pos = len(data1)
            data = io.BytesIO()

            # an entry wich consists of two items carries the position
            # of the second item in its header.
            data.write(struct.pack('<Q', pos))

            data.write(data1)
            data.write(data2)
            data = data.getvalue()

            # mark the entry as pair
            type_ = (type1, type2)

        else:
            type_, data = self._serialize_data(data)

        super().write(data, meta)
        self.cnt.info.append(type_)

    def read(self, index, n_samples):

        # get the entries as raw bytes from the superclass implementation
        entries = super().read(index, n_samples)

        res = []
        for i, entry in enumerate(entries):
            data, meta = entry
            type_ = self.cnt.info[index+i]

            if isinstance(type_, tuple):
                # If the type is a pair (x,y), deserialize independently
                buf = io.BytesIO(data)

                # First, get the position of the second item from the header
                pos, = struct.unpack('<Q', buf.read(8))

                # Read the first and second item
                data1 = buf.read(pos)
                data2 = buf.read()

                # Then deserialize the independently.
                data1 = self._deserialize(data1, type_[0])
                data2 = self._deserialize(data2, type_[1])

                res.append(((data1, data2), meta))
            else:
                data = self._deserialize(data, type_)
                res.append((data, meta))

        return res



from vergeml.utils import download_files
from vergeml.utils import Error
import shutil
import os.path
import zipfile
import tarfile


def download(env):
    """60000 tiny colour images in 10 classes.

The CIFAR-10 dataset consists of 60000 32x32 colour images in 10 
classes, with 6000 images per class. There are 50000 training 
images and 10000 test images.

Credits:
  Alex Krizhevsky
  Vinod Nair
  Geoffrey Hinton

For more information visit: 
https://www.cs.toronto.edu/~kriz/cifar.html"""

    url = "https://www.cs.toronto.edu/~kriz/cifar-10-python.tar.gz"

    samples_dir = env.get('base.samples_dir')

    print("Downloading CIFAR-10 to {}.".format(samples_dir))

    src_dir = download_files([url], dir=env.get('base.cache_dir'))
    path = os.path.join(src_dir, "cifar-10-python.tar.gz")

    tarf = tarfile.TarFile(path, 'r:gz')
    tarf.extractall(src_dir)
    tarf.close()

    shutil.rmtree(src_dir)

    print("Finished downloading CIFAR-10.")


download.__info__ = [
    ('Samples', '60000'),
    ('Type', 'Images'),
    ('Resolution', '32x32 RGB'),
    ('Size', '163 MB')
]



from vergeml.command import command, CommandPlugin
from vergeml.option import option
from vergeml.utils import VergeMLError
import os
import os.path
import subprocess
import re
import webbrowser

@command('tensorboard', descr="Run Tensorboard.")
@option('@AIs', type='List[@]', default=[])
@option('host', type='Optional[str]', descr="What host to listen to.")
@option('port', type=int, default=6006, descr="What port to serve Tensorboard on.")
class TensorboardService(CommandPlugin):

    def __call__(self, args, env):
        trainings_dir = env.get('trainings-dir')
        ais_with_tbstats = []

        if not os.path.exists(trainings_dir):
            raise VergeMLError("No trainings found.", "To run tensorboard, please train an AI first.")

        for dir in os.listdir(trainings_dir):
            if dir.startswith("."):
                continue
            stats_dir = os.path.join(trainings_dir, dir, "stats")

            if not os.path.exists(stats_dir):
                continue

            if not any(map(lambda d: d.startswith("events.out"), os.listdir(stats_dir))):
                continue

            ais_with_tbstats.append(dir)

        AIs = args["@AIs"] or ais_with_tbstats

        if not AIs:
            raise VergeMLError("No trainings found.", "To run tensorboard, please train an AI first.")

        for AI in AIs:
            if AI not in ais_with_tbstats:
                raise VergeMLError("Not tensorboard stats found for @{}".format(AI))

        dirs = []
        for AI in AIs:
            dirs.append(AI + ":" + os.path.join(trainings_dir, AI, "stats"))
        dirs = ",".join(dirs)
        cmd = ["tensorboard", "--logdir", dirs, "--port", str(args['port'])]
        if 'host' in args and args['host']:
            cmd.append('--host')
            cmd.append(args['host'])

        url = None

        try:
            for line in _run_command(cmd):
                line = line.decode('utf-8').rstrip()
                match = re.match(r".*?(http:[^ ]*)", line)
                if match:
                    url = match.group(1)
                    webbrowser.open(url)
                print(line)
        except FileNotFoundError:
            raise VergeMLError("Command 'tensorboard' not found.",
                               "Please install tensorboard (pip install tensorboard)")

def _run_command(command):
    p = subprocess.Popen(command,
                         stdout=subprocess.PIPE,
                         stderr=subprocess.STDOUT)
    return iter(p.stdout.readline, b'')


from vergeml.command import command, CommandPlugin
from vergeml.option import option

@command('plot', descr="Show different kinds of plots.")
@option('type', type=str, subcommand='vergeml.plot')
class PlotCommand(CommandPlugin):

    def __call__(self, args, env):
        plugin = self.plugins.get('vergeml.plot', args['type'])(args['type'])
        plugin(args, env)


from vergeml.plugins import PLUGINS
from vergeml.option import Option
from vergeml.command import command
from vergeml.utils import VergeMLError
from vergeml.display import DISPLAY
import urllib
import urllib.request
import os
import tempfile
import uuid
import time

def dataset(name, descr=None, long_descr=None):
    """Define a dataset.

    :param name:        Name of the dataset.
    :param descr:       A short description of the dataset
    :param long_descr:  A long description
    """
    return command(name=name, descr=descr, long_descr=long_descr)

class DatasetPlugin:
    def __init__(self):
        self.progress_bar = None
        self.progress = None
        self.start_time = None
    
    def __call__(self, args, env):
        raise NotImplementedError
    
    def download_files(self, urls, env, headers=None, dir=None):

        if dir is None:
            dir = env.get('cache-dir')
        
        dest_directory = os.path.join(dir, "tmp_" + str(uuid.uuid4()))
        
        if not os.path.exists(dest_directory):
            os.makedirs(dest_directory)

        for data_url in urls:
            if isinstance(data_url, tuple):
                data_url, download_file = data_url
            else:
                download_file = data_url.split('/')[-1]

            download_path = os.path.join(dest_directory, download_file)

            if headers:
                opener = urllib.request.build_opener()
                opener.addheaders = headers
                urllib.request.install_opener(opener)

            try:
                urllib.request.urlretrieve(data_url, filename=download_path, 
                                           reporthook=self._report_hook(download_file), data=None)
            except Exception as e:
                raise VergeMLError("Could not download {}: {}".format(data_url, e))
            finally:
                if headers:
                    urllib.request.install_opener(urllib.request.build_opener())

        return dest_directory

    def _report_hook(self, filename):   
        def _update_to(count, block_size, total_size):
            self.progress = (count, block_size, total_size)

            if total_size:
                if not self.progress_bar:
                    self.progress_bar = DISPLAY.progressbar(steps=total_size, title=filename, post=self._post)
                    self.progress_bar.start()
                
                self.progress_bar.update(count*block_size)
                if count * block_size +1 >= total_size:
                    self.progress_bar.stop()
                    self.progress_bar = None
        return _update_to
    
    def _post(self):
        count, block_size, total_size = self.progress
        if total_size:
            if count == 0:
                self.start_time = time.time()
                return ""
            
            duration = time.time() - self.start_time
            progress_size = int(count * block_size)
            if duration != 0.0:
                speed = int(progress_size / (1024 * duration))
                return " %d MB, %d KB/s" % ( progress_size / (1024 * 1024), speed)
            else:
                return ""
        else:
            return ""


from vergeml.plugins import PLUGINS
from vergeml.utils import VergeMLError
import os.path
import io
from vergeml.command import Command
from collections import OrderedDict
from typing import Optional
import json
import cgi
import tempfile
import shutil
import re

class WSGIApp:

    def __init__(self, env):
        self.env = env
        self.fns = OrderedDict()
        for model_fn in Command.find_functions(env.model_plugin):
            cmd = Command.discover(model_fn)
            if cmd.type == 'predict':
                self.fns[cmd.name] = (cmd, model_fn)

        if not len(self.fns):
            raise VergeMLError(f"@{env.trained_model} can't be run as a REST service.")

    def handler(self, environ, start_response):
        path_info = environ.get('PATH_INFO')
        method = environ.get('REQUEST_METHOD')

        status = '404 NOT FOUND'
        response_body = 'Not Found.'.encode('utf-8')
        content_type = 'text/plain'

        if path_info.lstrip("/") in self.fns.keys():
            if method == "GET":
                status, response_body, content_type = self._serve_function(path_info.lstrip("/"), environ)
            elif method == "POST":
                status, response_body, content_type = self._predict(path_info.lstrip("/"), environ)
        elif path_info == '/css/rest.css':
            status, response_body, content_type = self._serve_file('text/css', 'css', 'rest.css')
        elif path_info == '/img/logo.png':
            status, response_body, content_type = self._serve_file('image/png', 'img', 'logo.png')
        elif path_info == '/img/logo-big.png':
            status, response_body, content_type = self._serve_file('image/png', 'img', 'logo-big.png')
        elif path_info == '/js/rest.js':
            status, response_body, content_type = self._serve_file('application/javascript', 'js', 'rest.js')
        elif path_info == '/':
            status, response_body, content_type = self._serve_index(environ)

        response_headers = [
            ('Content-Type', content_type),
        ]

        start_response(status, response_headers)

        return [response_body]

    def _serve_file(self, mime_type, *path):
        path = list(path)
        path.insert(0, 'other')
        path.insert(0, os.path.dirname(__file__))

        with open(os.path.join(*path), 'rb') as f:
            return '200 OK', f.read(), mime_type

    def _serve_index(self, environ):
        index = _TEMPLATE.format(name=self.env.trained_model,
                                 content='<a id="hamburger">&#9776;</a> <div id="logo-big"> </div>',
                                 menu=self._make_menu())
        return '200 OK', index.encode('utf-8'), 'text/html'

    def _serve_function(self, name, environ):
        host = "http://" + environ.get('SERVER_NAME') + ":" + str(environ.get('SERVER_PORT'))
        content_a = _TEMPLATE_A.format(host=host,
                                        name=name,
                                        descr=self.fns[name][0].descr,
                                        fields=self._make_fields(name))
        index = _TEMPLATE.format(name=self.env.trained_model,
                                 content=content_a + _TEMPLATE_B,
                                 menu=self._make_menu(name))
        return '200 OK', index.encode('utf-8'), 'text/html'

    def _make_menu(self, active=None):
        res = []
        for k in self.fns.keys():
            klass = 'active' if k == active else ''
            res.append(f'<a class="{klass}" href="/{k}">/{k}</a>')

        return "\n".join(res)

    def _make_fields(self, name):
        cmd = self.fns[name][0]


        res = []
        for o in cmd.options:
            if o.is_at_option() or o.command_line:
                continue

            display_name = o.name.strip("<").strip(">")
            descr = o.descr + " (" + o.human_type() + ")"

            if o.has_type(float, int, 'Optional[float]', 'Optional[int]'):
                value = '' if o.default is None else o.default
                res.append(_TEMPLATE_NUMBER.format(name=display_name, value=value, descr=descr))
            elif o.has_type('File', 'Optional[File]'):
                res.append(_TEMPLATE_FILE.format(name=display_name, multiple="", descr=descr, label='Select file...'))
            elif o.has_type('List[File]'):
                res.append(_TEMPLATE_FILE.format(name=display_name, multiple="multiple", descr=descr, label='Select files...'))
            elif o.has_type(str) and isinstance(o.validate, (list, tuple)):
                res.append(_TEMPLATE_LIST(name=display_name, descr=descr, opts=o.validate, default=o.default))
            else:
                value = '' if o.default is None else o.default
                res.append(_TEMPLATE_STRING.format(name=display_name, value=value, descr=descr))
        return "\n".join(res)

    def _predict(self, name, environ):
        if not is_post_request(environ):
            return '400 BAD REQUEST', b'Bad Request', 'text/plain'

        cmd, fn = self.fns[name]
        form = get_post_form(environ)

        args = {}
        tempdirs = []
        try:
            for o in cmd.options:

                if o.is_at_option():
                    args[o.name] = self.env.trained_model
                    continue

                if o.command_line:
                    args[o.name] = o.default
                    continue

                form_name = o.name.lstrip("<").rstrip(">")

                if form_name in form:
                    value = form[form_name]

                    if o.has_type("File", "List[File]", "Optional[File]"):
                        files = []
                        if not isinstance(value, list):
                            value = [value]
                        for item in value:
                            if not item.filename or not item.file:
                                continue
                            name = secure_filename(item.filename)
                            dir = tempfile.mkdtemp()
                            tempdirs.append(dir)
                            name = os.path.join(dir, name)
                            with open(name, 'wb') as f:
                                shutil.copyfileobj(item.file, f)
                            files.append(name)

                        if o.has_type("File"):
                            if not files:
                                raise VergeMLError("Missing argument: {}".format(o.name))
                            args[o.name] = files[0]
                        elif o.has_type("Optional[File]"):
                            if files:
                                args[o.name] = files[0]
                        else:
                            args[o.name] = files
                    else:
                        args[o.name] = value.value

                    if o.name in args:
                        value = args[o.name]
                        value = o.cast_value(value)
                        value = o.transform_value(value)
                        o.validate_value(value)
                        args[o.name] = value
                else:
                    if not o.is_optional():
                        raise VergeMLError("Missing argument: {}".format(o.name))
                    elif o.default:
                        args[o.name] = o.default
        except VergeMLError as e:
            for d in tempdirs:
                shutil.rmtree(d)
            return '400 BAD REQUEST', e.message.encode('utf-8'), 'text/plain'

        try:
            res = fn(args, self.env)
        finally:
            for d in tempdirs:
                shutil.rmtree(d)

        return '200 OK', json.dumps(res).encode('utf-8'), 'application/json'

# https://wsgi.readthedocs.io/en/latest/specifications/handling_post_forms.html
def is_post_request(environ):
    if environ['REQUEST_METHOD'].upper() != 'POST':
        return False
    content_type = environ.get('CONTENT_TYPE', 'application/x-www-form-urlencoded')
    return (content_type.startswith('application/x-www-form-urlencoded')
        or content_type.startswith('multipart/form-data'))

def get_post_form(environ):
    assert is_post_request(environ)
    input = environ['wsgi.input']
    post_form = environ.get('wsgi.post_form')
    if (post_form is not None
        and post_form[0] is input):
        return post_form[2]
    # This must be done to avoid a bug in cgi.FieldStorage
    environ.setdefault('QUERY_STRING', '')
    fs = cgi.FieldStorage(fp=input,
                          environ=environ,
                          keep_blank_values=1)

    return fs

# from https://github.com/pallets/werkzeug/blob/master/werkzeug/utils.py
_filename_ascii_strip_re = re.compile(r'[^A-Za-z0-9_.-]')
_windows_device_files = ('CON', 'AUX', 'COM1', 'COM2', 'COM3', 'COM4', 'LPT1',
                         'LPT2', 'LPT3', 'PRN', 'NUL')
def secure_filename(filename):
    if isinstance(filename, str):
        from unicodedata import normalize
        filename = normalize('NFKD', filename).encode('ascii', 'ignore')
        filename = filename.decode('ascii')
    for sep in os.path.sep, os.path.altsep:
        if sep:
            filename = filename.replace(sep, ' ')
    filename = str(_filename_ascii_strip_re.sub('', '_'.join(
                   filename.split()))).strip('._')

    # on nt a couple of special files are present in each folder.  We
    # have to ensure that the target file is not such a filename.  In
    # this case we prepend an underline
    if os.name == 'nt' and filename and \
       filename.split('.')[0].upper() in _windows_device_files:
        filename = '_' + filename

    return filename

_TEMPLATE = """

<!DOCTYPE html>
<html>
<head>
<meta name="viewport" content="width=device-width, initial-scale=1">
<link rel="stylesheet" type="text/css" href="css/rest.css">

</head>
<body>
    <div id="overlay"></div>
    <div class="container">
        <div class="sidebar">
            <a href="/"><img src="img/logo.png" id="logo">@{name}</a>
            {menu}
        </div>
        <div class="content">
            {content}
        </div>
    </div>
</body>
<script src="js/rest.js"></script>
</html>

"""

_TEMPLATE_A = """
<div class="column" id="A">
    <div class="header">
        <a id="hamburger">&#9776;</a>
        <a class="method">POST</a>
        <span class="url"><span class="host">{host}</span>/{name}</span>
    </div>
    <div class="body">
        <form id="main-form" action="{name}" method="POST" enctype="multipart/form-data">
            <h2>{descr}</h2>

            {fields}

            <div class="submit">
                <input id="submit-button" type="submit" value="SEND">
            </div>
        </form>
    </div>
</div>
"""

_TEMPLATE_B = """
<div class="column" id="B">
    <div class="header">
        <span id="time" class="dark-box">
            TIME xxxms
        </span>
        <span id="status-ok" class="green-box">
            STATUS OK
        </span>
        <span id="status-err" class="red-box">
            STATUS ERROR
        </span>
    </div>
    <div class="body">
        <div class="json">
        </div>
    </div>
</div>
"""

_TEMPLATE_NUMBER = """
<div class="row">
    <div class="input">
        <span class="field">{name}</span>
        <input type="number" name="{name}" value="{value}">
    </div>
    <div class="description">{descr}</div>
</div>
"""


_TEMPLATE_STRING = """
<div class="row">
    <div class="input">
        <span class="field">{name}</span>
        <input type="text" name="{name}" value="{value}">
    </div>
    <div class="description">{descr}</div>
</div>
"""

_TEMPLATE_FILE = """
<div class="row">
    <div class="input">
        <span class="field">{name}</span>
        <input name="{name}" id="__{name}_input__" type="file" onchange="updateFileInput('{name}')" {multiple}>
        <label id="__{name}_label__" for="__{name}_input__">{label}</label>
    </div>
    <div class="description">{descr}</div>
</div>
"""

def _TEMPLATE_LIST(name, descr, opts, default):
    def _fopt(opt):
        sel = 'selected="selected"' if opt == default else ''
        return f'<option value="{opt}" {sel}>{opt}</option>'

    options = "\n".join(map(_fopt, opts))

    return """
    <div class="row">
        <div class="input">
            <span class="field">{name}</span>
            <select name="{name}">
                {options}
            </select>
        </div>
        <div class="description">{descr}</div>
    </div>
    """.format(name=name, descr=descr, options=options)


from vergeml.command import command, CommandPlugin
from vergeml.option import option
from vergeml.plots import load_labels

import os.path
import csv
from vergeml.utils import VergeMLError
import numpy as np

@command('confusion-matrix', descr="Plot a confusion matrix.")
@option('@AI')
@option('normalize', type='bool', default=False, descr="When true normalize the confusion matrix.", short="n", flag=True)
class ConfusionMatrixPlot(CommandPlugin):

    def __call__(self, args, env):
        # Plotting a confusion matrix needs the model to follow the convention
        # - labels.txt in checkpoints
        # - predictions.csv in stats

        import itertools
        import numpy as np
        import matplotlib.pyplot as plt
        from sklearn.metrics import confusion_matrix
        from vergeml.plots import load_labels, load_predictions

        try:
            labels = load_labels(env)
        except FileNotFoundError:
            raise VergeMLError("Can't plot confusion matrix - not supported by model.")

        nclasses = len(labels)

        try:
            y_test, y_score = load_predictions(env, nclasses)
        except FileNotFoundError:
            raise VergeMLError("Can't plot confusion matrix - not supported by model.")

        # From:
        # https://scikit-learn.org/stable/auto_examples/model_selection/plot_confusion_matrix.html#sphx-glr-auto-examples-model-selection-plot-confusion-matrix-py

        def plot_confusion_matrix(cm, classes, AI,
                                  normalize=False,
                                  cmap=plt.cm.Blues): # pylint: disable=E1101
            """
            This function prints and plots the confusion matrix.
            Normalization can be applied by setting `normalize=True`.
            """
            if normalize:
                cm = cm.astype('float') / cm.sum(axis=1)[:, np.newaxis]
                plt.title(f"Confusion matrix for @{AI} (normalized)")
            else:
                plt.title(f'Confusion matrix for @{AI}')

            plt.imshow(cm, interpolation='nearest', cmap=cmap)

            plt.colorbar()
            tick_marks = np.arange(len(classes))
            plt.xticks(tick_marks, classes, rotation=45)
            plt.yticks(tick_marks, classes)

            fmt = '.2f' if normalize else 'd'
            thresh = cm.max() / 2.
            for i, j in itertools.product(range(cm.shape[0]), range(cm.shape[1])):
                plt.text(j, i, format(cm[i, j], fmt),
                        horizontalalignment="center",
                        color="white" if cm[i, j] > thresh else "black")

            plt.ylabel('True label')
            plt.xlabel('Predicted label')
            plt.tight_layout()

        y_test = np.argmax(y_test, axis=1)
        y_pred = np.argmax(y_score, axis=1)
        cnf_matrix = confusion_matrix(y_test, y_pred)
        np.set_printoptions(precision=2)


        plt.figure()
        plot_confusion_matrix(cnf_matrix, normalize=args['normalize'], classes=labels, AI=args['@AI'])
        plt.show()



"""This module contains interfaces to the underlying data loading.
"""

from typing import List, Any, Union, Callable, Optional
import random

import numpy as np

from vergeml.utils import VergeMLError
from vergeml.views import BatchView, IteratorView
from vergeml.io import SourcePlugin
from vergeml.operation import BaseOperation
from vergeml.loader import FileCachedLoader, LiveLoader, MemoryCachedLoader
from vergeml.plugins import PLUGINS
from vergeml.utils import introspect
from vergeml.display import DISPLAY
from vergeml.utils import SPLITS


class Labels(list):
    """A list of labels.
    """
    pass


class BoundingBoxes(list):
    """A list of bounding boxes.
    """
    pass


class BoundingBox: # pylint: disable=R0903
    """A bounding box defined by coordinates and size.
    """
    def __init__(self, label: str, x: int, y: int, width: int, height: int): # pylint: disable=R0913

        # pylint: disable=C0103
        self.label = label
        self.x = x
        self.y = y
        self.width = width
        self.height = height


class Data:
    """Handle loading, augmentation and caching of your sample data.

    The easiest way to set up Data is by using it from the environment
    object.

        xy_train = env.data.load()
        # xy_train is now [(x1, y1), (x2, y2), ...]

    Data will be automatically set up by the environment and ready to
    use. Alternatively, Data can be set up manually by providing input,
    output and ops.
    """

    def __init__(self, # pylint: disable=R0913
                 env: 'Environment' = None,
                 input: SourcePlugin = None, # pylint: disable=W0622
                 output: SourcePlugin = None,
                 ops: List[BaseOperation] = None,
                 random_seed: int = 42,
                 cache_dir: str = '.cache',
                 cache_input: Union[str, bool] = 'mem',
                 cache_output: Union[str, bool] = False,
                 plugins=PLUGINS):

        """For automatic configuration, pass in an env object. To
        manually setup the data class, you need to provide the
        input (a SourcePlugin object).

        :param env: an Environment object used to set up Data. If an
                    environment is provided, other options are ignored.

        :param input: the input (SourcePlugin)

        :param output: the output (SourcePlugin)

        :param ops: a list of preprocessing operations (BaseOperation)

        :param random_seed: the random seed to use

        :param cache_input: config of input caching, default: False.
                            possible values: 'mem', 'disk' or False

        :param cache_output: config of output caching, default: 'disk'
        """

        self.cache_dir = cache_dir
        self.env = env
        self.input = input

        # output defaults to input
        self.output = output or input
        self.ops = (ops or []).copy()

        self.random_seed = random_seed
        self.cache_input = cache_input
        self.cache_output = cache_output

        self.plugins = plugins
        self.loader = None
        self._progress_bar = None


        if env:
            # When we are provided with an environment, set everything
            # up according to the environment's configuration.
            self._setup_from_env()
        else:

            # Sanity check
            assert cache_input in ('mem', 'disk', False)
            assert cache_output in ('mem', 'disk', False)
            assert self.input is not None

            self.loader = self._get_loader(cache_input, cache_output)

    def _get_loader(self, cache_input, cache_output):

        if cache_input in ('disk', 'mem'):

            # When doing input caching, wrap the input object in
            # a cached loader.

            loader_class = FileCachedLoader if cache_input == 'disk' else MemoryCachedLoader
            input_loader = loader_class(self.cache_dir, self.input)
            input_loader.progress_callback = self._progress_callback
        else:

            # otherwise, use the input object directly
            input_loader = self.input

        if cache_output in ('disk', 'mem'):

            # set up output caching

            loader_class = FileCachedLoader if cache_output == 'disk' else MemoryCachedLoader
            loader = loader_class(self.cache_dir, input_loader, self.ops, self.output)
            loader.progress_callback = self._progress_callback

            return loader

        return LiveLoader(self.cache_dir, input_loader, self.ops, self.output)

    @property
    def meta(self):
        """Sample metadata (e.g. labels).
        """
        self.loader.begin_read_samples()
        meta = self.loader.meta
        self.loader.end_read_samples()
        return meta

    def load(self, # pylint: disable=R0913
             split: str = 'train',
             view: str = 'list',
             layout: str = 'tuples',
             batch_size: int = 64,
             fetch_size: Optional[int] = None,
             infinite: bool = False,
             with_meta: bool = False,
             randomize: bool = False,
             transform_x: Callable[[Any], Any] = lambda x: x,
             transform_y: Callable[[Any], Any] = lambda y: y):

        """
        :param split: The split to load. One of "train", "val", "test".

        :param view: How to return the data. Option are:

                    - "list" (default): reads all data into memory and
                      return it as python list or optionally as numpy
                      array (when layout is set to 'arrays').

                    - "batch": return a generator, splitting the data
                      into batches of batch_size. The returned object
                      supports getting the length (number of batches)
                      via the len() function.

                    - "iter": return the data as an **iterator** object.

        :param layout: Determines how x, y and (optionally meta) is
                       returned.
                       Can be one of:

                       - "tuples": shape the data as (x,y) pairs.

                         [(x1,y1), (x2,y2), ...]

                       - "lists": shape x,y as distinct lists [xs], [ys]

                         ([x1, x2, x3], [y1, y2, y3])

                       - "arrays": return numpy arrays [xs], [ys]

                         array([[1, 2, 3],
                                [4, 5, 6]])

        :param batch_size: Sample batch size. Applies to "batch" view
                           only. Defaults to 64.

        :param fetch_size: Fetch samples in pairs of fetch_size. None
                           means the system will automatically set a
                           fetch size.

        :param infinite: Applies to "batch" and "iter" views. If set to
                         True, the returned object will be an infinite
                         generator object.

            NOTE for KERAS users: This setting is useful when used in
            with model.fit_generator(). Since len() will return the
            number of steps per epoch, steps_per_epoch can be left
            unspecified when calling fit_generator().

        :param with_meta: If True, will return meta in addition to x
                          and y.

        :param randomize: If True, the data will be returned in random
                          order.

        :param transform_x: a function that takes x as an argument and
                            returns a transformed version.
                            Defaults to no transformation.

        :param transform_y: a function that takes x as an argument and
                            returns a transformed version.
                            Defaults to no transformation."""

        assert view in ('list', 'batch', 'iter')
        assert layout in ('tuples', 'lists', 'arrays')
        fetch_size = fetch_size or 8

        if view == 'list':
            return _load_list(loader=self.loader,
                              split=split,
                              layout=layout,
                              with_meta=with_meta,
                              randomize=randomize,
                              random_seed=self.random_seed,
                              transform_x=transform_x,
                              transform_y=transform_y)

        if view == 'batch':

            return BatchView(loader=self.loader,
                             split=split,
                             layout=layout,
                             batch_size=batch_size,
                             fetch_size=fetch_size,
                             infinite=infinite,
                             with_meta=with_meta,
                             randomize=randomize,
                             random_seed=self.random_seed,
                             transform_x=transform_x,
                             transform_y=transform_y)

        if view == 'iter':
            return IteratorView(loader=self.loader,
                                split=split,
                                fetch_size=fetch_size,
                                infinite=infinite,
                                with_meta=with_meta,
                                randomize=randomize,
                                random_seed=self.random_seed,
                                transform_x=transform_x,
                                transform_y=transform_y)


        # never runs in this code, make the linter happy
        assert False
        return None

    def _base_env_config(self):
        """Get the base configuration values from env (splits etc.)
        """
        keys = ('val-split', 'test-split', 'samples-dir', 'random-seed', 'trainings-dir')
        return {k: self.env.get(k) for k in keys}

    def _setup_input(self):
        """Set up input from env.
        """

        # get the name of the input plugin
        input_name = self.env.get('data.input.type')
        if not input_name:
            raise VergeMLError("data.input.type is not defined.")

        # get input configuration and merge base config
        input_conf = self.env.get('data.input').copy()
        input_conf.update(self._base_env_config())

        # instantiate the input plugin
        input_class = self.plugins.get('vergeml.io', input_name)
        if not input_class:
            raise VergeMLError("input name not found: {}".format(input_name))

        # TODO validate configuration and set defaults
        del input_conf['type']

        self.input = input_class(input_conf)

    def _setup_ops(self):
        """Set up ops from env.
        """

        # set up preprocessing operations
        self.ops = []

        for conf in self.env.get('data.preprocess') or []:
            if isinstance(conf, str):
                conf = dict(name=conf)
            else:
                conf = conf.copy()

            # every preprocessing operations needs a name property
            name = conf.get('op', None)
            if not name:
                raise VergeMLError("Name missing in data.preprocess item.")
            del conf['op']

            # instantiate the preprocessing plugin
            plugin = self.plugins.get('vergeml.operation', name)
            if not plugin:
                raise VergeMLError("preprocess plugin not found: {}".format(name))

            # check arguments
            intro = introspect(plugin)
            mandatory = set(intro.args[1:]).difference(set(intro.defaults.keys()))
            missing = set(mandatory).difference(conf.keys())
            unknown = set(conf.keys()).difference(intro.args[1:])

            # TODO type checking

            # report missing or unknown arguments
            if missing:
                msg = "preprocess operation {} is missing argument(s): {}"
                raise VergeMLError(msg.format(name, missing))

            if unknown:
                msg = "preprocess operation {} received unknown argument(s): {}"
                raise VergeMLError(msg.format(name, unknown))

            operation = plugin(**conf)
            self.ops.append(operation)

    def _setup_output(self):
        """Set up output from env.
        """

        # get the name of the output plugin or set it to input
        output_name = self.env.get('data.output.type') or self.env.get('data.input.type')

        # get output configuration or set it to input configuration and merge
        # with base config
        output_conf = self.env.get('data.output') or self.env.get('data.input').copy()
        output_conf['name'] = output_name
        output_conf.update(self._base_env_config())

        # instantiate the output plugin
        output_class = self.plugins.get('vergeml.io', output_name)
        if not output_class:
            raise VergeMLError("output name not found: {}".format(output_name))

        self.output = output_class(output_conf)


    def _setup_cache(self):
        """Set up caching from env.
        """

        cache = self.env.get("data.cache")
        if cache == 'mem-in':
            self.cache_input, self.cache_output = 'mem', False
        elif cache == 'disk-in':
            self.cache_input, self.cache_output = 'disk', False
        elif cache == 'mem':
            self.cache_input, self.cache_output = False, 'mem'
        elif cache in ('disk', 'auto'):
            self.cache_input, self.cache_output = False, 'disk'
        elif cache == 'none':
            self.cache_input, self.cache_output = False, False


    def _setup_from_env(self):
        """Configure using the environment object.
        """

        self.random_seed = self.env.get('random-seed')
        self.cache_dir = self.env.cache_dir()

        self._setup_input()
        self._setup_ops()
        self._setup_output()
        self._setup_cache()

        self.loader = self._get_loader(self.cache_input, self.cache_output)

    def num_samples(self, split):
        """Return the number of samples in split.
        """

        self.loader.begin_read_samples()
        res = self.loader.num_samples(split)
        self.loader.end_read_samples()
        return res

    def _progress_callback(self, current, total):
        if total:

            if current == -1:

                if self.cache_input == 'mem':
                    msg = "Caching input samples in memory ..."
                elif self.cache_output == 'mem':
                    msg = "Caching output samples in memory ..."
                elif self.cache_input == 'disk':
                    msg = "Caching input samples on disk ..."
                elif self.cache_output == 'disk':
                    msg = "Caching output samples on disk ..."
                else:
                    msg = None

                if msg:
                    DISPLAY.print(msg)
            else:
                if not self._progress_bar:
                    # start the progress bar
                    self._progress_bar = DISPLAY.progressbar(steps=total,
                                                             label="samples", keep=False)
                    self._progress_bar.start()

                self._progress_bar.update(current)

                if current + 1 == total:
                    # close the progress bar on the last step
                    self._progress_bar.stop()
                    print("")



def _load_list(loader,  # pylint: disable=R0914,R0913
               random_seed,
               split,
               transform_x,
               transform_y,
               with_meta,
               randomize,
               layout):
    # pylint: disable=C0103
    res = []
    loader.begin_read_samples()

    num_samples = loader.num_samples(split)

    for sample in loader.read_samples(split, 0, num_samples):
        x, y, m = sample.x, sample.y, sample.meta
        x, y = transform_x(x), transform_y(y)
        if with_meta:
            res.append((x, y, m))
        else:
            res.append((x, y))

    loader.end_read_samples()

    if randomize:
        random.Random(random_seed).shuffle(res)

    if layout in ('lists', 'arrays'):
        res = tuple(map(list, zip(*res)))
        if not res:
            res = ([], [], []) if with_meta else ([], [])

    if layout == 'arrays':
        xs, ys, *meta = res
        res = tuple([np.array(xs), np.array(ys)] + meta)
    return res



from vergeml.img import ImageType
from vergeml.operation import OperationPlugin, operation
from vergeml.option import option
from PIL import Image
from vergeml.utils import VergeMLError

@operation('flip-vertical', topic="image", descr="Vertically flip and image.")
@option('chance', validate='>=0, <=1', default=1.)
class FlipVerticalOperation(OperationPlugin):

    def __init__(self, chance:float=1.0, apply=None):
        super().__init__(apply)
        self.chance = chance
    

    def transform_xy(self, x, y, rng):
      
        if rng.uniform(0.0, 1.0) < self.chance:

            x = x if not isinstance(x, ImageType) else x.transpose(Image.FLIP_TOP_BOTTOM)
            y = y if not isinstance(y, ImageType) else y.transpose(Image.FLIP_TOP_BOTTOM)
            
        return x, y



"""
This module implements the data structures returned by Data.load() to
support the unique data loading requirements of different deep learning
libraries.
"""

import random
import itertools
from typing import Callable, Any

import numpy as np

def _rand_batch_ixs(num_samples: int, batch_size: int, fetch_size: int, random_seed: int):
    """A generator which yields a list of tuples (offset, size) in random order.

    This list will be used by the data loader to efficiently load samples and pass it to
    the model during training.

    :param num_samples: Number of available samples.
    :param batch_size: The size of the batch to fill.
    :param fetch_size: Desired fetch_size.
    :param random_seed: RNG seed.
    """
    rng = random.Random(random_seed)
    batch, batch_count = [], 0

    while True:
        if fetch_size * 3 < num_samples:
            # if the number of samples is too small, having a random offset
            # makes no sense
            offset = rng.randint(0, fetch_size)
        else:
            offset = 0

        ixs = list(range(offset, num_samples - offset, fetch_size))
        rng.shuffle(ixs)

        # collect enough samples to fill the batch

        while ixs:
            next_fetch = ixs.pop(0)

            # calculate the next fetch size depending on the samples remaining
            # and the number of samples required to fill the batch

            next_fetch_size = min(fetch_size,
                                  num_samples - next_fetch, batch_size - batch_count)

            batch.append((next_fetch, next_fetch_size))
            batch_count += next_fetch_size

            if batch_count == batch_size:
                yield batch
                batch, batch_count = [], 0

def _ser_batch_ixs(num_samples, batch_size):
    """A generator which yields a list of tuples (offset, size) in serial order.

    :param num_samples: Number of available samples.
    :param batch_size: The size of the batch to fill.
    """
    current_index = 0
    batch, batch_count = [], 0

    while True:
        next_fetch = current_index
        next_fetch_size = min(batch_size - batch_count, num_samples - next_fetch)

        batch.append((next_fetch, next_fetch_size))
        batch_count += next_fetch_size

        if batch_count == batch_size:

            # If we have enough samples to fill the batch size, yield
            # the indices and reset the batch count.
            yield batch
            batch, batch_count = [], 0

        current_index += next_fetch_size

        if current_index == num_samples:
            current_index = 0



def _pumpfn(ix_gen):
    while True:
        yield from next(ix_gen, None)

class BatchView: # pylint: disable=R0902
    """Generator that returns data as batches (optionally infinite).
    """

    def __init__(self, # pylint: disable=R0913
                 loader,
                 split,
                 layout: str = 'tuples',
                 batch_size: int = 64,
                 fetch_size: int = 8,
                 infinite: bool = False,
                 with_meta: bool = False,
                 randomize: bool = False,
                 random_seed: int = 42,
                 transform_x: Callable[[Any], Any] = lambda x: x,
                 transform_y: Callable[[Any], Any] = lambda y: y):

        self.loader = loader
        self.split = split

        self.loader.begin_read_samples()
        num_samples = self.loader.num_samples(self.split)
        self.loader.end_read_samples()

        self.infinite = infinite
        self.with_meta = with_meta
        self.transform_x = transform_x
        self.transform_y = transform_y
        self.layout = layout
        self.num_batches = num_samples // batch_size
        self.current_batch = 0

        if randomize:
            ix_fn = lambda: _rand_batch_ixs(num_samples, batch_size, fetch_size, random_seed)
        else:
            ix_fn = lambda: _ser_batch_ixs(num_samples, batch_size)

        # We generate two identical ix generators - one for the view and
        # one for the loader
        self.ix_gen = ix_fn()
        self.loader.pump(self.split, _pumpfn(ix_fn()))

    def __iter__(self):
        self.current_batch = 0
        return self

    def __len__(self):
        return self.num_batches

    def __next__(self):

        if self.current_batch >= self.num_batches and not self.infinite:
            raise StopIteration

        # BEGIN loading samples from the data loader
        self.loader.begin_read_samples()

        res = []
        for index, n_samples in next(self.ix_gen):

            samples = self.loader.read_samples(self.split, index, n_samples)
            for sample in samples:

                # pylint: disable=C0103
                x, y, m = sample.x, sample.y, sample.meta
                x, y = self.transform_x(x), self.transform_y(y)
                res.append((x, y, m) if self.with_meta else (x, y))

        self.loader.end_read_samples()
        # END loading samples

        self.current_batch += 1

        # rearrange the result according to the configured layout
        if self.layout in ('lists', 'arrays'):
            res = tuple(map(list, zip(*res)))

        if self.layout == 'arrays':

            # pylint: disable=C0103
            xs, ys, *meta = res
            res = tuple([np.array(xs), np.array(ys)] + meta)

        return res

class IteratorView: # pylint: disable=R0902
    """Generator that returns one sample at a time.
    """
    def __init__(self, # pylint: disable=R0913
                 loader,
                 split,
                 fetch_size=8,
                 infinite=False,
                 with_meta=False,
                 randomize=False,
                 random_seed=42,
                 transform_x=lambda x: x,
                 transform_y=lambda y: y):

        self.loader = loader
        self.split = split

        self.loader.begin_read_samples()
        self.num_samples = self.loader.num_samples(self.split)
        self.loader.end_read_samples()

        self.infinite = infinite
        self.with_meta = with_meta
        self.transform_x = transform_x
        self.transform_y = transform_y
        self.fetch_size = fetch_size
        self.rng = random.Random(random_seed) if randomize else None
        self.ixs = None

        self.current_index = 0
        self._shuffle()

    def _shuffle(self):
        self.ixs = range(0, self.num_samples)

        if self.rng:
            # When randomizing sample order, make sure to lay out samples
            # according to fetch size to improve performance.
            self.ixs = [self.ixs[i:i + self.fetch_size]
                        for i in range(0, len(self.ixs), self.fetch_size)]
            self.rng.shuffle(self.ixs)
            self.ixs = list(itertools.chain.from_iterable(self.ixs))

    def __iter__(self):
        return self

    def __len__(self):
        return self.num_samples

    def __next__(self):

        if self.current_index >= self.num_samples:
            self.current_index = 0
            self._shuffle()
            if not self.infinite:
                raise StopIteration

        index = self.ixs[self.current_index]
        sample = self.loader.read_samples(self.split, index, 1)[0]

        # pylint: disable=C0103
        x, y, m = sample.x, sample.y, sample.meta
        x, y = self.transform_x(x), self.transform_y(y)
        res = (x, y, m) if self.with_meta else (x, y)

        self.current_index += 1
        return res



from vergeml.utils import download_files
from vergeml.utils import Error
import shutil
import os.path
import zipfile
import tarfile
import urllib.request


def download(env):
    """Large-scale CelebFaces Attributes.

CelebFaces Attributes Dataset (CelebA) is a large-scale face 
attributes dataset with more than 200K celebrity images.

Credits:
   Ziwei Liu   
   Ping Luo   
   Xiaogang Wang   
   Xiaoou Tang
   Multimedia Laboratory, The Chinese University of Hong Kong 

For more information visit:
http://mmlab.ie.cuhk.edu.hk/projects/CelebA.html"""

    samples_dir = env.get('base.samples_dir')

    print("Downloading celeba to {}.".format(samples_dir))

    url = "https://drive.google.com/uc?id=0B7EVK8r0v71pZjFTYXZWM3FlRnM&export=download"

    token = None
    fp = urllib.request.urlopen(url)
    info = fp.info()
    fp.close()

    cookies = info.get("Set-Cookie")

    for cookie in cookies.split(";"):
        if cookie.startswith("download_warning"):
            _, token = cookie.split("=")
            token = token.strip()
            break

    if not token:
        raise Error("Could not get google drive download token.")

    url = "https://drive.google.com/uc?export=download&confirm={}&id=0B7EVK8r0v71pZjFTYXZWM3FlRnM".format(token)

    src_dir = download_files([(url, "img_align_celeba.zip")], headers=[('Cookie', cookies)], dir=env.get('base.cache_dir'))

    path = os.path.join(src_dir, "img_align_celeba.zip")

    print("Extracting data.")

    zipf = zipfile.ZipFile(path, 'r')
    zipf.extractall(src_dir)
    zipf.close()

    for src_file in os.listdir(os.path.join(src_dir, "img_align_celeba")):
        if src_file.startswith("."):
            continue

        shutil.copy(os.path.join(src_dir, "img_align_celeba", src_file), samples_dir)

    shutil.rmtree(src_dir)

    print("Finished downloading celeba.")


download.__info__ = [
    ('Samples', '202,599'),
    ('Type', 'Labeled Images'),
    ('Size', '1.44 GB')
]




from vergeml.utils import VergeMLError
from vergeml.dataset import DatasetPlugin, dataset
from vergeml.display import DISPLAY
import shutil
import os.path
import zipfile
import tarfile

_LONG_DESCR = """Stanford dogs dataset for fine-grained visual categorization.

The Stanford Dogs dataset contains images of 120 breeds of dogs 
from around the world. This dataset has been built using images 
and annotation from ImageNet for the task of fine-grained image 
categorization.

Credits:
  Aditya Khosla, Nityananda Jayadevaprakash, Bangpeng Yao and 
  Li Fei-Fei. Novel dataset for Fine-Grained Image Categorization. 
  First Workshop on Fine-Grained Visual Categorization (FGVC), IEEE 
  Conference on Computer Vision and Pattern Recognition (CVPR), 2011. 
  http://people.csail.mit.edu/khosla/papers/fgvc2011.pdf

  J. Deng, W. Dong, R. Socher, L.-J. Li, K. Li and L. Fei-Fei, 
  ImageNet: A Large-Scale Hierarchical Image Database. IEEE Computer 
  Vision and Pattern Recognition (CVPR), 2009.
  http://www.image-net.org/papers/imagenet_cvpr09.pdf

For more information visit: 
http://vision.stanford.edu/aditya86/ImageNetDogs/""".strip()

_INFO = [
    ['Samples', '20,580'],
    ['Type', 'Labeled Images'],
    ['Size', '757 MB'],
]

_LONG_DESCR += "\n\n" + DISPLAY.table(_INFO, separate='none').getvalue()

_URL = "http://vision.stanford.edu/aditya86/ImageNetDogs/images.tar"

@dataset('dogs', descr="Stanford dogs dataset for fine-grained visual categorization.", long_descr=_LONG_DESCR)
class DogsDataset(DatasetPlugin):

    def __call__(self, args, env):
        samples_dir = env.get('samples-dir')

        for label in ['affenpinscher', 'afghan-hound', 'african-hunting-dog', 'airedale', 'american-staffordshire-terrier', 'appenzeller', 'australian-terrier', 'basenji', 'basset', 'beagle', 'bedlington-terrier', 'bernese-mountain-dog', 'black-and-tan-coonhound', 'blenheim-spaniel', 'bloodhound', 'bluetick', 'border-collie', 'border-terrier', 'borzoi', 'boston-bull', 'bouvier-des-flandres', 'boxer', 'brabancon-griffon', 'briard', 'brittany-spaniel', 'bull-mastiff', 'cairn', 'cardigan', 'chesapeake-bay-retriever', 'chihuahua', 'chow', 'clumber', 'cocker-spaniel', 'collie', 'curly-coated-retriever', 'dandie-dinmont', 'dhole', 'dingo', 'doberman', 'english-foxhound', 'english-setter', 'english-springer', 'entlebucher', 'eskimo-dog', 'flat-coated-retriever', 'french-bulldog', 'german-shepherd', 'german-short-haired-pointer', 'giant-schnauzer', 'golden-retriever', 'gordon-setter', 'great-dane', 'great-pyrenees', 'greater-swiss-mountain-dog', 'groenendael', 'ibizan-hound', 'irish-setter', 'irish-terrier', 'irish-water-spaniel', 'irish-wolfhound', 'italian-greyhound', 'japanese-spaniel', 'keeshond', 'kelpie', 'kerry-blue-terrier', 'komondor', 'kuvasz', 'labrador-retriever', 'lakeland-terrier', 'leonberg', 'lhasa', 'malamute', 'malinois', 'maltese-dog', 'mexican-hairless', 'miniature-pinscher', 'miniature-poodle', 'miniature-schnauzer', 'newfoundland', 'norfolk-terrier', 'norwegian-elkhound', 'norwich-terrier', 'old-english-sheepdog', 'otterhound', 'papillon', 'pekinese', 'pembroke', 'pomeranian', 'pug', 'redbone', 'rhodesian-ridgeback', 'rottweiler', 'saint-bernard', 'saluki', 'samoyed', 'schipperke', 'scotch-terrier', 'scottish-deerhound', 'sealyham-terrier', 'shetland-sheepdog', 'shih-tzu', 'siberian-husky', 'silky-terrier', 'soft-coated-wheaten-terrier', 'staffordshire-bullterrier', 'standard-poodle', 'standard-schnauzer', 'sussex-spaniel', 'tibetan-mastiff', 'tibetan-terrier', 'toy-poodle', 'toy-terrier', 'vizsla', 'walker-hound', 'weimaraner', 'welsh-springer-spaniel', 'west-highland-white-terrier', 'whippet', 'wire-haired-fox-terrier', 'yorkshire-terrier']:
            dest = os.path.join(samples_dir, label)
            if os.path.exists(dest):
                raise VergeMLError("Directory {} already exists in samples dir: {}".format(label, dest))

        print("Downloading Stanford dogs to {}.".format(samples_dir))

        src_dir = self.download_files([_URL], env=env, dir=env.get('cache-dir'))
        path = os.path.join(src_dir, "images.tar")

        print("Extracting data...")

        tarf = tarfile.TarFile(path, 'r')
        tarf.extractall(src_dir)
        tarf.close()

        for src_file in os.listdir(os.path.join(src_dir, "Images")):
            if src_file.startswith("."):
                continue

            _, dest_file = src_file.split("-", maxsplit=1)
            dest_file = dest_file.replace("_", "-")
            dest_file = dest_file.lower()
            shutil.copytree(os.path.join(src_dir, "Images", src_file),
                            os.path.join(samples_dir, dest_file))

        shutil.rmtree(src_dir)

        print("Finished downloading Stanford dogs.")



from vergeml.img import INPUT_PATTERNS, open_image, fixext
from vergeml.io import source, SourcePlugin, Sample
import random
import numpy as np
from PIL import Image
import os.path

@source('image', descr="Load image files.", input_patterns=INPUT_PATTERNS)
class ImageSource(SourcePlugin):
    input_patterns = INPUT_PATTERNS

    def __init__(self, config: dict={}):
        self.files = None
        super().__init__(config)
            
    def begin_read_samples(self):
        if self.files:
            return
        
        self.files = self.scan_and_split_files()

    def num_samples(self, split: str) -> int:
        return len(self.files[split])

    def read_samples(self, split, index, n=1):
        items = self.files[split][index:index+n]
        items = [(open_image(filename), meta) for filename, meta in items]
        
        res = []
        for img, meta in items:
            rng = random.Random(str(self.random_seed) + meta['filename'])
            res.append(Sample(img, None, meta.copy(), rng)) 

        return res

    def transform(self, sample):
        sample.x = np.asarray(sample.x)
        sample.y = None
        return sample

    def hash(self, state: str) -> str:
        return super().hash(state + self.hash_files(self.files))
    
    def supports_preview(self):
        return True
    
    def write_preview(self, output_dir: str, split: str, sample: Sample):
        filename = sample.meta['filename']
        
        name = fixext(os.path.basename(filename), sample.x)
        split_dir = os.path.join(output_dir, split)
        if not os.path.exists(split_dir):
            os.makedirs(split_dir)

        path = self.preview_filename(os.path.join(split_dir, name))
        sample.x.save(path)
        return path



"""Parse configuration files.
"""

from copy import deepcopy
import re

import yaml

from vergeml.utils import VergeMLError, did_you_mean
from vergeml.plugins import PLUGINS
from vergeml.io import Source
from vergeml.operation import Operation

def parse_device(section, device_id=None, device_memory=None):
    """Parse the device section of the config file.
    """
    section = deepcopy(section or {})

    # As a shortcut, device can be set directly, e.g. device: gpu:1
    # This is equivalent to saying
    # device:
    #   id: gpu:1

    if isinstance(section, str):
        section = {'id': section}

    # If --device is set from the command line, it overrides the YAML file.
    if device_id:
        section['id'] = device_id

    # Override --memory flag
    if device_memory:
        section['memory'] = device_memory

    res = {
        'id': 'auto',
        'memory': 'auto',
        'grow-memory': True
    }

    # First, raise an error if an unknown option is encountered
    _raise_unknown_option('device', res.keys(), section.keys(), 'device')

    _parse_device_id(res, section)

    _parse_device_memory(res, section)

    _parse_device_grow_memory(res, section)

    return res

def _parse_device_id(res, section):
    """Parse the id option in the device section.
    """
    if 'id' in section:

        value = section['id'].strip()

        if not re.match(r"^(gpu:[0-9]+|gpu|cpu|auto)", value):
            raise _invalid_option('device.id', 'device')

        if value == 'gpu':
            value = 'gpu:0'

        res['id'] = value

def _parse_device_memory(res, section):
    """Parse the memory option in the device section.
    """
    if 'memory' in section:

        value = section['memory'].strip()

        if isinstance(value, float):
            if value < 0. or value > 1.:
                raise _invalid_option('device.memory', 'device')

            res['memory'] = value


        if value != 'auto':
            if not re.match(r'^[0-9]+(\.[0-9]*)?%$', value):
                raise _invalid_option('device.memory', 'device')

            try:
                value = float(value.rstrip('%'))
            except ValueError:
                raise _invalid_option('device.memory', 'device')

            if value < 0. or value > 100.:
                raise _invalid_option('device.memory', 'device')

            res['memory'] = value/100

def _parse_device_grow_memory(res, section):
    """Parse the grow-memory option in the device section.
    """
    if 'grow-memory' in section:

        try:
            value = bool()
        except ValueError:
            raise _invalid_option('device.grow-memory', 'device')

        res['grow-memory'] = value


def parse_data(section, cache=None, plugins=PLUGINS):
    """Parse the data section of the config file.
    """
    section = deepcopy(section or {})

    # There are several shortcuts when setting up data.

    # First, one can just specify the input plugin as a value of the data option
    if isinstance(section, str):
        section = {'input': {'type': section}}

    # input and output can be set directly too
    for k in ('input', 'output'):
        if isinstance(section.get(k), str):
            section[k] = {'type': section[k]}

    # Override --cache flag
    if cache:
        section['cache'] = cache

    res = {
        'cache': 'auto',
        'preprocess': []
    }

    # Raise an error if an unknown option is encountered
    _raise_unknown_option('data', ('input', 'output', 'cache', 'preprocess'),
                          section.keys(), 'data')

    _parse_data_cache(res, section)

    _parse_data_source(res, section, 'input', plugins)

    _parse_data_source(res, section, 'output', plugins)

    _parse_data_preprocess(res, section, plugins)

    return res


_VALID_CACHE_VALUES = ('none', 'mem', 'disk', 'mem-in', 'disk-in', 'auto')
def _parse_data_cache(res, section):

    if 'cache' in section:
        value = section['cache']

        if not value in _VALID_CACHE_VALUES:

            suggestion = did_you_mean(_VALID_CACHE_VALUES, value)
            raise _invalid_option('data.cache', help_topic='cache', suggestion=suggestion)
        res['cache'] = value


def _parse_data_source(res, section, key, plugins):

    if key in section:
        source_section = section[key]

        # Type must be set explicitly when input or output is specified in the config file.
        if not 'type' in source_section:
            raise VergeMLError(f"Missing option 'type'.",
                               suggestion='Please specify the {key} type of your data',
                               help_topic='data', hint_type='key', hint_key=f'data.{key}')

        type_ = source_section['type']
        res[key] = {'type': type_}

        # Find the source definition
        plugin = plugins.get("vergeml.io", type_)

        if not plugin:
            suggestion = did_you_mean(plugins.keys('vergeml.io'), type_)
            raise _invalid_option(f"data.{key}.type",
                                  help_topic='data',
                                  suggestion=suggestion)

        source = Source.discover(plugin)

        # check for unknown options
        _raise_unknown_option(f'data.{key}', map(lambda o: o.name, source.options),
                              set(source_section.keys()) - set(['type']), 'data')

        options = list(filter(lambda o: o.name != 'type', source.options))

        # validate and transform each option
        for option in options:
            try:

                if option.name in source_section:
                    value = source_section[option.name]
                    value = option.cast_value(value)
                    value = option.transform_value(value)
                    option.validate_value(value)
                    res[key][option.name] = value

                # deal with missing options
                elif option.is_required():
                    raise VergeMLError(f"Missing option '{option.name}'.",
                                       suggestion=f'Please add the missing option.',
                                       help_topic='data', hint_type='key', hint_key=f'data.{key}')
                elif option.default:
                    res[key][option.name] = option.default


            except VergeMLError as err:
                err.hint_key = f'data.{key}.{option.name}'
                raise err




def _parse_data_preprocess(res, section, plugins):
    if 'preprocess' in section:

        pre_section = section['preprocess']

        if not isinstance(pre_section, list):
            message = "Invalid value - must be a list of preprocess operations."
            raise VergeMLError(message,
                               "Please fix the entry in the project file.",
                               help_topic="preprocess",
                               hint_type='key',
                               hint_key='data.preprocess.')

        operations = []
        for index, config in enumerate(pre_section):
            if not isinstance(config, dict):
                raise VergeMLError("Invalid entry in preprocess - must be key value pairs.",
                                   "Please fix the entry in the project file.",
                                   help_topic="preprocess",
                                   hint_type='key',
                                   hint_key='data.preprocess.' + str(index))
            elif not 'op' in config:
                raise VergeMLError("Invalid entry in preprocess - missing 'op' key.",
                                   "Please fix the entry in the project file.",
                                   help_topic="preprocess",
                                   hint_type='key',
                                   hint_key='data.preprocess.' + str(index))
            op_name = config['op']
            plugin = plugins.get("vergeml.operation", op_name)
            if not plugin:
                raise VergeMLError(f"Invalid entry in preprocess - unknown operation '{op_name}'.",
                                   "Please fix the entry in the project file.",
                                   help_topic="preprocess",
                                   hint_type='value',
                                   hint_key=f"data.preprocess.{index}.op")


            operation = Operation.discover(plugin)
            # options = list(filter(lambda o: o.name != 'op', operation.options))#

            # check for unknown options
            _raise_unknown_option(f'data.preprocess.{index}',
                                  [o.name for o in operation.options],
                                  set(config.keys()) - set(['op']), 'preprocess')

            opdict = {'op': op_name}
            for option in operation.options:
                try:
                    if option.name in config:
                        value = config[option.name]

                        value = option.cast_value(value)
                        value = option.transform_value(value)
                        option.validate_value(value)

                        opdict[option.name] = value

                    # deal with missing options
                    elif option.is_required():
                        raise VergeMLError(f"Missing option '{option.name}'.",
                                           suggestion=f'Please add the missing option.',
                                           help_topic='data', hint_type='key',
                                           hint_key=f'data.preprocess.{index}')
                    elif option.default:
                        opdict[option.name] = option.default

                except VergeMLError as err:
                    err.hint_key = f'data.preprocess.{index}.{option.name}'
                    raise err


            operations.append(opdict)
        res['preprocess'] = operations


def parse_command(command, section):
    """Parse a command in the config file.
    """
    section = section or {}
    res = {}
    subk = []

    sub_option = next(filter(lambda c: c.subcommand, command.options), None)

    if sub_option:
        subk = list(command.plugins.keys(sub_option.subcommand))

    _raise_unknown_option(command.name, [o.name for o in command.options] + subk,
                          section.keys(), command.name)

    for option in command.options:

        if option.name in section:
            value = section[option.name]
            value = option.cast_value(value)
            value = option.transform_value(value)
            option.validate_value(value)
            res[option.name] = value

    return res


def _raise_unknown_option(key, valid, options, help_topic):

    unknown = list(set(options) - set(valid))

    if bool(unknown):
        first = unknown[0]
        suggestion = did_you_mean(valid, first)
        raise _invalid_option(f'{key}.{first}', help_topic=help_topic,
                              suggestion=suggestion, kind='key')



def _invalid_option(key, help_topic=None, suggestion=None, kind='value'):
    label = "Invalid value for option" if kind == 'value' else "Invalid option"
    return VergeMLError(f"{label} '{key}'.", suggestion, help_topic=help_topic,
                        hint_type=kind, hint_key=key)



def load_yaml_file(filename, label='config file', loader=yaml.Loader):
    """Load a yaml config file.
    """
    try:
        with open(filename, "r") as file:
            res = yaml.load(file.read(), Loader=loader) or {}
            if not isinstance(res, dict):
                msg = f"Please ensure that {label} consists of key value pairs."
                raise VergeMLError(f"Invalid {label}: {filename}", msg)
            return res
    except yaml.YAMLError as err:
        if hasattr(err, 'problem_mark'):
            mark = getattr(err, 'problem_mark')
            problem = getattr(err, 'problem')
            message = f"Could not read {label} {filename}:"
            message += "\n" + display_err_in_file(filename, mark.line, mark.column, problem)
        elif hasattr(err, 'problem'):
            problem = getattr(err, 'problem')
            message = f"Could not read {label} {filename}: {problem}"
        else:
            message = f"Could not read {label} {filename}: YAML Error"

        suggestion = f"There is a syntax error in your {label} - please fix it and try again."

        raise VergeMLError(message, suggestion)

    except OSError as err:
        msg = "Please ensure the file exists and you have the required access privileges."
        raise VergeMLError(f"Could not open {label} {filename}: {err.strerror}", msg)


class _YAMLAnalyzer(yaml.reader.Reader, yaml.scanner.Scanner):

    def __init__(self, stream):
        yaml.reader.Reader.__init__(self, stream)
        yaml.scanner.Scanner.__init__(self)

def _get_location(ann, key, kind):
    mark = ann.get_mark()

    # try to mark the value
    if kind == 'value':
        if isinstance(ann.peek_token(), yaml.ValueToken):
            # forward to the next token
            tok = ann.get_token()
            if isinstance(ann.peek_token(), yaml.ScalarToken):
                # forward again
                tok = ann.get_token()
                length = len(tok.value)
                return (mark.line, mark.column + 1, length)

    length = len(key) + 1
    return (mark.line, max(0, mark.column - length), length)

def yaml_find_definition(stream, key, kind='key'): # pylint: disable=R0912
    """Find the location of the definition of key in the YAML source.
    """
    assert kind in ('key', 'value')
    keys = list(map(lambda k: int(k) if k.isdigit() else k, key.split(".")))
    level = -1
    matches = [False] * len(keys)
    indices = []

    ann = _YAMLAnalyzer(stream)

    tok = ann.get_token()
    while tok: # pylint: disable=R1702
        if isinstance(tok, (yaml.BlockMappingStartToken, yaml.BlockSequenceStartToken)):
            level += 1
            indices.append(0)
        elif isinstance(tok, yaml.ValueToken) and \
             isinstance(ann.peek_token(), yaml.BlockEntryToken):
            level += 1
            indices.append(0)
            # this is a special case since no start and end tokens are emitted for a simple list
            tok = ann.get_token()
            while tok:
                if isinstance(tok, yaml.BlockEntryToken):
                    if 0 <= level < len(keys) and isinstance(keys[level], int):
                        if keys[level] == indices[level] and all(matches[:level]):
                            matches[level] = True
                            if all(matches):
                                return _get_location(ann, str(keys[-1]), kind)
                    indices[level] += 1
                elif not isinstance(tok, (yaml.ScalarToken, yaml.ValueToken)):
                    break

        elif isinstance(tok, yaml.BlockEndToken):
            level -= 1
            indices.pop()
        elif isinstance(tok, yaml.KeyToken) and 0 <= level < len(keys) and \
             isinstance(keys[level], str):
            name = ann.get_token().value

            if name == keys[level] and all(matches[:level]):
                matches[level] = True
                if all(matches):
                    return _get_location(ann, keys[-1], kind)

        elif isinstance(tok, yaml.BlockEntryToken):

            if 0 <= level < len(keys) and isinstance(keys[level], int):
                if keys[level] == indices[level] and all(matches[:level]):
                    matches[level] = True
                    if all(matches):
                        return _get_location(ann, keys[-1], kind)

            indices[level] += 1

        tok = ann.get_token()
    return None

def display_err_in_file(filename, line, column, message, length=1, nlines=3): # pylint: disable=R0913
    """Mark the source code error location in file and return a string for display"""
    with open(filename, "r") as file:
        return _display_err(filename, line, column, message, length, nlines, file.read())


def _display_err(filename, line, column, message, length, nlines, content): # pylint: disable=R0913
    """Mark the source code error location in content and return a string for display"""
    lines = content.splitlines()
    start = max(0, line+1-nlines)
    res = [f"File {filename}, line {line+1}:{column+1}"]
    res.append(str('-' * (len(res[0]) + 7)))
    res += lines[start:line+1]
    res += [(' ' * column) + ("^" * length), message]
    return "\n".join(res)



from pkg_resources import iter_entry_points, EntryPoint
from vergeml.utils import VergeMLError
import re

class _PluginManager:

    def get(self, group, name):
        raise NotImplementedError
    
    def set(self, group, name, plugin):
        raise NotImplementedError
    
    def keys(self, group):
        raise NotImplementedError
    
    def all(self, group):
        return {k:self.get(group, k) for k in self.keys(group)}

class _SetupToolsPluginManager(_PluginManager):

    def __init__(self):
        self.__plugins = {}

    def get(self, group, name):
        self.__plugins.setdefault(group, {})
        if name not in self.__plugins[group]:
            ep = next(iter_entry_points(group, name=name), None)
            if ep:
                self.__plugins[group][name] = ep.resolve()
        return self.__plugins[group].get(name, None)

    
    def set(self, group, name, plugin):
        self.__plugins.setdefault(group, {})
        self.__plugins[group][name] = plugin

    
    def keys(self, group):
        self.__plugins.setdefault(group, {})
        entry_points = list(iter_entry_points(group, name=None))
        ep_keys = set(map(lambda ep: ep.name, entry_points))
        return set(self.__plugins[group].keys()).union(ep_keys)
    
class _DictPluginManager(_PluginManager):

    def __init__(self):
        self.__plugins = {}

    def get(self, group, name):
        self.__plugins.setdefault(group, {})
        return self.__plugins[group].get(name, None)
    
    def set(self, group, name, plugin):
        self.__plugins.setdefault(group, {})
        self.__plugins[group][name] = plugin

    def keys(self, group):
        self.__plugins.setdefault(group, {})
        return set(self.__plugins[group].keys()) 


PLUGINS = _SetupToolsPluginManager()






"""
A collection of classes implementing various data loading techniques.
"""

import operator
import os.path
import threading
import queue

from functools import reduce
from typing import List

from vergeml.io import Sample
from vergeml.utils import SPLITS
from vergeml.cache import MemoryCache, SerializedFileCache

class _Pump(threading.Thread):
    """Continuously perform data loading in a background thread like a
    pump.

    Pump continuously loads samples and fills up a queue, which is then
    read back by the loader. To determine which samples to read, it
    takes an infinite generator function as an argument. The generator
    returns the index and the number of samples to read next.
    Once max_items are waiting in the queue, it will pause loading
    samples.
    """

    def __init__(self, loader, split, ix_gen, max_items):
        """
        :param loader: Object responsible for loading samples.
        :param split: train, val or test.
        :param ix_gen: An infinite generator yielding tuples (ix, n).
        :param max_items: The maximum number of samples to have in the
                          queue.
        """
        super().__init__()
        self.ix_gen = ix_gen
        self.outq = queue.Queue(max_items)
        self.split = split
        self.loader = loader
        self.daemon = True


    def run(self):
        self.loader.begin_read_samples()

        while True:
            index, n_samples = next(self.ix_gen)
            samples = self.loader.perform_read(self.split, index, n_samples)

            # Blocks when the queue is full (until samples are read).
            self.outq.put((index, n_samples, samples))

    def perform_read(self, _split: str, index: int, n_samples: int = 1):
        """Read samples from the queue which were previously read in the
        background thread.

        Samples must be read in the exact order in which they were
        placed in the queue, i.e. index and n_samples must match.
        """


        # remove samples from the queue. (if there are no samples available,
        # block until the background thread puts samples in the queue)
        index_, n_samples_, samples = self.outq.get()

        # Sanity check
        assert index_ == index
        assert n_samples_ == n_samples

        return samples

    def __del__(self):
        # (try) to clean up...
        self.loader.end_read_samples()


class Loader:
    """Abstract base class for data loaders.
    """

    def __init__(self, cache_dir, input, ops=None, output=None, transform=True): # pylint: disable=W0622
        self.cache_dir = cache_dir
        self.input = input
        self.ops = ops or []
        self.output = output
        self.cache = {}
        self.pumps = {}

        self._progress_callback = lambda n, t: None

    @property
    def meta(self):
        """Metadata of the loaded samples, e.g. all labels.
        """
        return self.input.meta

    @property
    def progress_callback(self):
        """Set this callback in order to receive progress updates.
        """
        return self._progress_callback

    @progress_callback.setter
    def progress_callback(self, value):
        """Setter for progress_callback.
        """
        self._progress_callback = value


    def pump(self, split, ix_gen, max_items=100):
        """Set up the pump for a split.

        Typically used by a view to set up the pumping mechanism. This
        will start a background thread which will continuously load
        samples, also during training.

        It works by receiving a generator which provides the pump with
        the index and number of the next samples to load. When
        read_samples is called later, samples must be read in the same
        order as previously returned by the generator.
        """
        if not split in self.pumps:
            self.pumps[split] = _Pump(self, split, ix_gen, max_items)
            self.pumps[split].start()

    def begin_read_samples(self):
        """Prepare to start reading samples.
        """
        raise NotImplementedError

    def num_samples(self, split: str) -> int:
        """Get the number of samples in split.
        """
        return len(self.cache[split])

    def read_samples(self, split: str, index: int, n_samples: int = 1) -> Sample:
        """Read n_samples starting at index from the cache.
        """
        samples = []

        # either read from a pump or directly from self.
        reader = self.pumps.get(split, self)

        for item in reader.perform_read(split, index, n_samples):
            x, y = item[0] # pylint: disable=C0103

            meta, rng = item[1]
            samples.append(Sample(x, y, meta, rng))

        return samples

    def perform_read(self, split: str, index: int, n_samples: int = 1):
        """Perform the actual read operation on the cache object.
        """
        return self.cache[split].read(index, n_samples)

    def end_read_samples(self):
        """Every call to begin_read_samples is closed with end_read_samples.
        """
        pass

    def _calculate_num_samples(self, split):
        """Calculate the total number of samples after applying ops.
        """
        num_samples = self.input.num_samples(split)

        # calculate how much the samples will be augmented after going through ops
        multiplier = reduce(operator.mul, map(lambda op: _get_multiplier(split, op), self.ops), 1)

        return int(num_samples * multiplier)

    def _calculate_hashed_state(self):
        """Get a hash representing the set of samples and the configuration.
        """

        # construct a string representing input configuration
        input_conf_str = str(sorted(self.input.configuration().items()))
        state = self.input.__class__.__name__ + input_conf_str

        if self.output:
        # construct a representation of ops and output configuration
            ops_state = "-".join([str(sorted(op.configuration().items())) for op in self.ops])
            out_state = self.output.__class__.__name__ + \
                str(sorted(self.output.configuration().items()))

            # and append it to state
            state = "-".join([state, ops_state, out_state])

        # input will handle hashing the state of sample data
        return self.input.hash(state)

    def _iter_samples(self, split, raw=False):
        """Iterate samples possibly applying operations.
        """

        num_samples = self.input.num_samples(split)
        readfn = lambda i: self.input.read_samples(split, i)[0]
        opfn = lambda s: [s]
        tffn = lambda s: s

        # apply operations
        if self.ops:
            op1, *oprest = self.ops
            opfn = lambda s: op1.process(s, oprest)

        if raw and not self.output:
        # read raw samples
            readfn = lambda i: self.input.read_raw_samples(split, i)[0]
        elif self.output:
        # transform the sample to output
            tffn = self.output.transform

        for index in range(num_samples):
            sample = readfn(index)
            for sample_ in opfn(sample):
                yield tffn(sample_)


def _get_multiplier(split, operation):
    has_split = hasattr(operation, 'apply') and operation.apply.intersection(set(SPLITS))

    if has_split and split not in operation.apply:
        return 1.0

    return operation.multiplier()

class MemoryCachedLoader(Loader):
    """Load sample data into a memory cache.
    """

    def begin_read_samples(self):
        if self.cache:
            return

        self.input.begin_read_samples()

        # copy meta
        if self.output:
            self.output.meta = self.input.meta

        self.cache = {k:MemoryCache() for k in SPLITS}
        total = sum(map(self._calculate_num_samples, SPLITS))

        i = 0
        self._progress_callback(-1, total)
        for split in SPLITS:
            cache = self.cache[split]
            for sample in self._iter_samples(split):
                cache.write((sample.x, sample.y), (sample.meta, sample.rng))
                self._progress_callback(i, total)
                i = i + 1

        self.input.end_read_samples()



class FileCachedLoader(Loader):
    """Cache sample data in a file cache.
    """

    def begin_read_samples(self):
        if self.cache:
            return

        self.input.begin_read_samples()

        # copy meta
        if self.output:
            self.output.meta = self.input.meta

        hashed_state = self._calculate_hashed_state()

        paths = [(split, self._cache_path(split, hashed_state)) for split in SPLITS]

        # get the total number of samples
        total = sum([self._calculate_num_samples(s) for s, p in paths
                     if not os.path.exists(p)])

        if total:
            i, cache = 0, None

            try:
                self._progress_callback(-1, total)

                for split, path in paths:
                    if not os.path.exists(path):
                        # we compress output data since its likely to be numpy arrays
                        cache = SerializedFileCache(path, "w", compress=bool(self.output))

                        for sample in self._iter_samples(split, raw=True):
                            cache.write((sample.x, sample.y), (sample.meta, sample.rng))
                            self._progress_callback(i, total)
                            i += 1
                        cache.close()

                        cache = None
            except (KeyboardInterrupt, SystemExit, Exception) as exc:
                # clean up in case of an exception

                # first, in case there is an open cache, close it
                if cache:
                    cache.close()

                # then, delete all cached files
                for _, path in paths:
                    if os.path.exists(path):
                        try:
                            os.unlink(path)
                        except: # pylint: disable=W0702
                            pass

                raise exc

        compress = bool(self.output)
        self.cache = {split:SerializedFileCache(path, "r", compress=compress)
                      for split, path in paths}
        self.input.end_read_samples()



    def read_samples(self, split: str, index: int, n_samples: int = 1) -> List[Sample]:
        samples = super().read_samples(split, index, n_samples)
        if not self.output:
            samples = [self.input.recover_raw_sample(sample) for sample in samples]
        return samples

    def _cache_path(self, split, hashed_state):
        return os.path.join(self.cache_dir, "{}-{}.cache".format(hashed_state, split))


class LiveLoader(Loader):
    """Load live sample data without caching.
    """

    multipliers = None
    rngs = None
    transform = True

    def begin_read_samples(self):
        if self.cache:
            return

        self.input.begin_read_samples()
         # copy meta
        if self.output:
            self.output.meta = self.input.meta

        self.multipliers = {}
        self.rngs = {}

        def _mul(split):
            return reduce(operator.mul, map(lambda op: _get_multiplier(split, op), self.ops), 1)

        for split in SPLITS:
            self.multipliers[split] = _mul(split)
            self.cache[split] = self._calculate_num_samples(split)
            self.rngs[split] = self.cache[split] * [None]

        self.input.end_read_samples()



    def num_samples(self, split: str) -> int:
        return self.cache[split]

    def perform_read(self, split: str, index: int, n_samples: int = 1): # pylint: disable=R0914

        mul = self.multipliers[split]
        offset = int(index % mul)
        start_index = int(index/mul)
        end_index = int((index+n_samples)/mul)
        read = max(1, int(n_samples/mul) + int(min(1, index%mul)))

        res = []

        samples = self.input.read_samples(split, start_index, read)
        if self.output and self.ops:
            op1, *oprest = self.ops

            for sample in samples:
                res.extend(op1.process(sample, oprest))

        else:
            res = samples

        if self.output and self.transform:
            res = [self.output.transform(sample) for sample in res]

        for sample, i in zip(res, range(start_index, end_index)):
            if self.rngs[split][i] is None:
                self.rngs[split][i] = sample.rng
            else:
                sample.rng = self.rngs[split][i]

        res = res[offset: offset+n_samples]

        return list(map(lambda s: ((s.x, s.y), (s.meta, s.rng)), res))



from vergeml.img import ImageType
from vergeml.operation import OperationPlugin, operation
from vergeml.option import option
from PIL import Image
from vergeml.utils import VergeMLError
import math

@operation('crop', topic="image", descr="Crop a region of an image at a fixed position.")
@option('x', type='Optional[int]', descr="x coordinate of the rectangle.", validate='>0')
@option('y', type='Optional[int]', descr="y coordinate of the rectangle.", validate='>0')
@option('position', type='Optional[str]', descr="Position of the rectangle",
        validate=("top-left", "top-right", "bottom-left", "bottom-right", "center"), default="center")
@option('width', type=int, descr="Width of the rectangle.", validate='>0')
@option('height', type=int, descr="Height of the rectangle.", validate='>0')
class CropOperation(OperationPlugin):
    type = ImageType

    def __init__(self, width:int, height:int, x:int = None, y:int = None, position:str="center", apply=None):

        super().__init__(apply)

        if bool(x) ^ bool(y):
            raise VergeMLError("Must specify both x and y when using absolute coordinates")

        VALID_POSITIONS = ("top-left", "top-right", "bottom-left", "bottom-right", "center")
        
        if not position in VALID_POSITIONS:
            raise VergeMLError("position must be one of: {}".format(", ".join(VALID_POSITIONS)))

        self.width = width
        self.height = height
        self.x = x
        self.y = y
        self.position = position
    
    def transform(self, img, rng):
        width, height = img.size

        if width < self.width:
            raise VergeMLError("Can't crop sample with width {} to {}.".format(width, self.width))
        
        if height < self.height:
            raise VergeMLError("Can't crop sample with height {} to {}.".format(height, self.height))

        if self.x or self.y:

            if width < self.width + self.x:
                raise VergeMLError("Can't crop sample with width {} to {} from x {}.".format(width, self.width, self.x))
            if height < self.height + self.y:
                raise VergeMLError("Can't crop sample with height {} to {} from y {}.".format(height, self.height, self.y))
            
            x = self.x
            y = self.y
        
        elif self.position == "top-left":
            x, y = 0,0
        elif self.position == "top-right":
            x, y = width - self.width, 0
        elif self.position == "bottom-left":
            x, y = 0, height - self.height
        elif self.position == "bottom-right":
            x, y = width - self.width, height - self.height
        elif self.position == "center":
            x, y = math.floor(width/2 - self.width/2), math.floor(height/2 - self.height/2)
        
        params = x, y, x + self.width, y + self.height
        return img.crop(params)


from vergeml.img import ImageType
from vergeml.operation import OperationPlugin, operation
from vergeml.option import option
from PIL import Image
from vergeml.utils import VergeMLError

@operation('random-crop', topic="image", descr="Crop random regions of an image.")
@option('width', type=int, descr="Width of the rectangle.", validate='>0')
@option('height', type=int, descr="Height of the rectangle.", validate='>0')
class RandomCropOperation(OperationPlugin):
    type = ImageType

    def __init__(self, width:int, height:int, apply=None):
        super().__init__(apply)
        self.width = width
        self.height = height

    def transform_xy(self, x, y, rng):
        imgs = [img for img in (x,y) if isinstance(img, ImageType)]

        if not len(imgs):
            raise VergeMLError("random_crop needs samples of type image")

        maxwidth = min([img.size[0] for img in imgs])
        maxheight = min([img.size[1] for img in imgs])

        if maxwidth < self.width:
            raise VergeMLError("Can't crop sample with width {} to {}.".format(maxwidth, self.width))
        
        if maxheight < self.height:
            raise VergeMLError("Can't crop sample with height {} to {}.".format(maxheight, self.height))

        maxx = maxwidth - self.width
        maxy = maxheight - self.height

        xco = rng.randint(0, maxx)
        yco = rng.randint(0, maxy)
        params = xco, yco, xco + self.width, yco + self.height

        if isinstance(x, ImageType):
            x = x.crop(params)
        
        if isinstance(y, ImageType):
            y = y.crop(params)

        return x, y


"""Preprocess command.
"""
import os.path
from vergeml.utils import VergeMLError, SPLITS
from vergeml.loader import LiveLoader
from vergeml.command import command
from vergeml.option import option
from vergeml.command import CommandPlugin

# pylint: disable=R0903

@command('preprocess', descr="Preprocess samples and save the output. ")
@option('<directory>', type='str', descr="Output directory.", default="pre")
@option('split', type='str', descr="The split to process.", default="all",
        validate=('train', 'val', 'test', 'all'))
@option('num-samples', type='Optional[int]', default=None,
        descr="The number of samples to process.", short='n')
class PreprocessCommand(CommandPlugin):
    """Preprocess samples and save the output.
    """

    def __call__(self, args, env):

        output_dir = args['<directory>']

        if not os.path.exists(output_dir):
            os.makedirs(output_dir)

        res = _preview(env.data, output_dir, args['split'], args['num-samples'])

        if not res:
            raise VergeMLError("Command preprocess not supported.")


def _preview(data, output_dir, split, num_samples):
    if not data.output.supports_preview():
        return False

    data.output.begin_preview(output_dir)

    loader = LiveLoader(data.cache_dir, data.input, data.ops, output=data.output)
    loader.transform = False
    loader.begin_read_samples()
    total = 0

    for split_ in SPLITS:
        if split not in ('all', split_):
            continue

        mul = int(loader.multipliers[split_])

        for index in range(loader.num_samples(split_)):
            if num_samples is not None and total >= num_samples:
                break

            samples = loader.read_samples(split_, index, mul)

            for sample in samples:
                data.output.write_preview(output_dir, split_, sample)

            total += 1


    loader.end_read_samples()
    data.output.end_preview(output_dir)

    return True



from vergeml.command import command, CommandPlugin
from vergeml.option import option
from vergeml.plots import load_labels

import os.path
import csv
from vergeml.utils import VergeMLError
import numpy as np

@command('pr', descr="Plot a precision/recall curve.")
@option('@AI')
@option('class', type='str', descr="The positive class.")

# TODO add plots for multi label
class PRPlot(CommandPlugin):

    def __call__(self, args, env):

        import numpy as np
        import matplotlib.pyplot as plt
        from sklearn.metrics import average_precision_score
        from sklearn.metrics import precision_recall_curve
        from vergeml.plots import load_labels, load_predictions

        try:
            labels = load_labels(env)
        except FileNotFoundError:
            raise VergeMLError("Can't plot PR curve - not supported by model.")

        nclasses = len(labels)
        if args['class'] not in labels:
            raise VergeMLError("Unknown class: " + args['class'])

        try:
            y_test, y_score = load_predictions(env, nclasses)
        except FileNotFoundError:
            raise VergeMLError("Can't plot PR curve - not supported by model.")

        # From:
        # https://scikit-learn.org/stable/auto_examples/model_selection/plot_precision_recall.html#sphx-glr-auto-examples-model-selection-plot-precision-recall-py

        ix = labels.index(args['class'])
        y_test = y_test[:,ix].astype(np.int)
        y_score = y_score[:,ix]

        precision, recall, _ = precision_recall_curve(y_test, y_score)
        average_precision = average_precision_score(y_test, y_score)

        plt.step(recall, precision, color='b', alpha=0.2, where='post')
        plt.fill_between(recall, precision, alpha=0.2, color='b', step='post')

        plt.xlabel('Recall ({})'.format(args['class']))
        plt.ylabel('Precision ({})'.format(args['class']))
        plt.ylim([0.0, 1.05])
        plt.xlim([0.0, 1.0])
        plt.title('Precision-Recall curve for @{0}: AP={1:0.2f}'.format(args['@AI'], average_precision))
        plt.show()



PARAM_DESCR = {
    'epochs': 'How many epochs to train.',
    'learning rate': 'Optimizer learning rate.',
    'batch size': 'The number of samples to use in one training batch.',
    'decay': 'Learning rate decay.',
    'early stopping': 'Early stopping delta and patience.',
    'dropout': 'Dropout rate.',
    'layers': 'The number of hidden layers.',
    'optimizer': 'Which optimizer to use.'
}

LONG_DESCR = {
    'learning rate': 'A hyperparameter which determines how quickly new learnings override old ones during training. In general, find a learning rate that is low enough that the network will converge to something useful, but high enough that the training does not take too much time.',

    'batch size': 'Defines the number of samples to be propagated through the network at once in a batch. The higher the batch size, the more memory you will need.',

    'epochs': 'The number of epochs define how often the network will see the complete set of samples during training.',

    'decay': 'When using learning rate decay, the learning rate is gradually reduced during training which may result in getting closer to the optimal performance.',

    'early stopping': 'Early Stopping is a form of regularization used to avoid overfitting when training. It is controlled via two parameters: Delta defines the minimum change in the monitored quantity to qualify as an improvement. Patience sets the number of epochs with no improvement after which training will be stopped.',

    'dropout': 'Dropout is a regularization technique for reducing overfitting in neural networks, The term "dropout" refers to dropping out units (both hidden and visible) in a neural network during training.',

    'layers': 'Deep learning models consist of a number of layers which contain one or more neurons. Typically, the neurons in one layer are connected to the next. Models with a higher number of layers can learn more complex representations, but are also more prone to overfitting.',

    'optimizer': 'The algorithm which updates model parameters such as the weight and bias values when training the network.',

    'sgd': 'Stochastic gradient descent, also known as incremental gradient descent, is a method for optimizing a neural network. It is called stochastic because samples are selected randomly instead of as a single group, as in standard gradient descent.',

    'adam': 'The Adam optimization algorithm is an extension to stochastic gradient descent which computes adaptive learning rates for each parameter. Adam is well suited for many practical deep learning problems.',

    'environment variables': """\
The following environment variables are available in VergeML:
- VERGEML_PERSISTENCE The number of persistent instances
- VERGEML_THEME The theme of the command line application
- VERGEML_FUNKY_ROBOTS Funky robots""",

    "overfitting": "When a model performs well on the training samples, but is not able to generalize well to unseen data. During training, a typical sign for overfitting is when your validation loss goes up while your training loss goes down.",

    "underfitting": "When a model can neither model the training data nor generalize to new data.",

    "hyperparameters": "Hyperparameters are the parameters of a model that can be set from outside, i.e. are not learned during training. (e.g. learning rate, number of layers, kernel size).",

    "random seed": "An integer value that seeds the random generator to generate random values. It is used to repeatably reproduce tasks and experiments.",

    "project": "A VergeML project is just a directory. Typically it contains a vergeml.yaml file, a trainings directory and a samples directory.",

    'project file': "A YAML file you can use to configure models, device usage, data processing and taks options.",

    'checkpoint': 'A checkpoint is a static image of a trained AI. It can be used to restore the AI after training and make predictions.',

    'stats': 'Stats are used to measure the performance of a model (e.g. accuracy).',

    'samples': 'Samples are pieces of data (e.g. images, texts) that is being used to train models to create AIs.',

    'val split': "Samples different from training samples that are used to evaluate the performance of model hyperparameters. You can set it via the --val-split option. See 'ml help split'.",

    'test split': "Samples different from training samples that are used to evaluate the final performance of the model. You can set it via the --test-split option. See 'ml help split'.",

    'split': 'split is a part of the sample data reserved for validation and testing (--val-split and --test-split options). It can be configured as either a percentage value (e.g. --val-split=10%) to reserve a fraction of training samples, a number to reserve a fixed number of samples, or a directory where the samples of the split are stored.',

    'cache dir': 'A directory which contains the processed data.',

}

SYNONYMS = {
    'stochastic gradient descent': 'sgd',
    'hyperparameter': 'hyperparameters',
    'project dir': 'project',
    'training samples': 'samples',
    'overfit': 'overfitting',
    'underfit': 'underfitting'
}

def long_descr(key):
    key = key.replace("-", " ")
    key = SYNONYMS.get(key, key)
    return LONG_DESCR.get(key, "").strip()

def short_param_descr(key):
    key = key.replace("-", " ")
    key = SYNONYMS.get(key, key)
    return PARAM_DESCR.get(key, "").strip()




"""Environment and related classes.
"""

import os.path
import datetime
import time
import inspect
import csv
from copy import deepcopy

import yaml
import numpy as np

from vergeml.utils import VergeMLError, did_you_mean, parse_split
from vergeml.utils import dict_get_path, dict_set_path

from vergeml.plugins import PLUGINS
from vergeml.data import Data, Labels
from vergeml.random_robot import random_robot_name, ascii_robot
from vergeml.libraries import KerasLibrary, TensorFlowLibrary, TorchLibrary
from vergeml.libraries import NumPyLibrary, PythonInterpreter
from vergeml.display import DISPLAY, TrainingFeedback
from vergeml.config import yaml_find_definition, display_err_in_file, load_yaml_file
from vergeml.config import parse_data, parse_device


# Represent labels as list
yaml.add_representer(Labels, lambda dump, dat: dump.represent_list(dat))

# Do not use aliases.
yaml.Dumper.ignore_aliases = lambda *_args: True

ENV = None

_DEFAULT_STATS = [dict(name='acc', title='Accuracy', category="TRAINING", format='.4f',
                       smooth=True, log=True),
                  dict(name='loss', title='Loss', category="TRAINING", format='.4f',
                       smooth=True, log=True),
                  dict(name='val-acc', title='Accuracy', category="VALIDATION", format='.4f',
                       smooth=False, log=True),
                  dict(name='val-loss', title='Loss', category="VALIDATION", format='.4f',
                       smooth=False, log=True),
                  dict(name='test-acc', title='Accuracy', category="TESTING", format='.4f',
                       smooth=False, log=False),
                  dict(name='test-loss', title='Loss', category="TESTING", format='.4f',
                       smooth=False, log=False),]

class Environment:

    _current_command = None

    @property
    def current_command(self):
        """The name of the currently executing command"""
        return self._current_command

    @current_command.setter
    def current_command(self, current_command):
        self._current_command = current_command

    def __init__(self, # pylint: disable=R0913,R0914
                 model=None,
                 project_file=None,
                 samples_dir=None,
                 test_split=None,
                 val_split=None,
                 cache_dir=None,
                 random_seed=None,
                 trainings_dir=None,
                 project_dir=None,
                 cache=None,
                 device=None,
                 device_memory=None,
                 trained_model=None,
                 is_global_instance=False,
                 # config=None,
                 plugins=PLUGINS,
                 display=DISPLAY):
        """Configure, train and save the results.

        :param model:           Name of the model plugin.
        :param project_file:    Optional path to the project file.
        :param samples_dir:     The directory where samples can be found. [default: samples]
        :param test_split:      The test split. [default: 10%]
        :param val_split:       The val split. [default: 10%]
        :param cache_dir:       The directory used for caching [default: .cache]
        :param random_seed:     Random seed. [default 42]
        :param trainings_dir:   The directory to save training results to. [default: trainings]
        :param project_dir:     The directory of the project. [default: current directory]
        :param trained_model:   Optional name of a trained model.
        :is_global_instance:    If true, this env can be accessed under the global var env.ENV.
                                [default: false]
        :config:                Additional configuration to pass to env, i.e. if not using a project
                                file.
        """

        super().__init__()

        # when called from the command line, we need to have a global instance
        if is_global_instance:
            global ENV # pylint: disable=W0603
            ENV = self

        # setup the display
        self.display = display

        # set the name of the trained model if given
        self.trained_model = trained_model

        # this holds the model object (not the name of the model)
        self.model_plugin = None
        # The results class (responsible for updating data.yaml with the latest results
        # during training)
        self.results = None
        # When a training is started, this holds the object responsible for coordinating the
        # training.
        self.training = None
        # Hold a proxy to the data loader.
        self._data = None

        self.plugins = plugins

        # Set up the base options from constructor arguments.
        self._config = {
            'samples-dir': samples_dir,
            'test-split': test_split,
            'val-split': val_split,
            'cache-dir': cache_dir,
            'random-seed': random_seed,
            'trainings-dir': trainings_dir,
            'model': model
        }

        # merge project file
        if project_file:
            doc = self._load_yaml_and_configure(project_file, 'project file',
                                                cache, device, device_memory)

            # the project file DOES NOT override values passed to the environment
            for k, val in doc.items():

                if not k in self._config or self._config[k] is None:
                    self._config[k] = val

        # after the project file is loaded, fill missing values
        project_dir = project_dir or ''
        defaults = {
            'samples-dir': os.path.join(project_dir, "samples"),
            'test-split': '10%',
            'val-split': '10%',
            'cache-dir': os.path.join(project_dir, ".cache"),
            'random-seed': 42,
            'trainings-dir': os.path.join(project_dir, "trainings"),
        }

        for k, val in defaults.items():
            if self._config[k] is None:
                self._config[k] = val

        self._validate_split(project_dir)

        # Have the data_file variable in outer scope for later when reporting errors
        data_file = None

        if self.trained_model:
            data_file = self._load_trained_model()

        self._load_model_plugin(project_file, data_file, model is not None)

        # always set up numpy and python
        self.configure('python')
        self.configure('numpy')

    def _validate_split(self, project_dir):
        """Validate the split configuration for val and test.
        """

        for split in ('val-split', 'test-split'):
            spltype, splval = parse_split(self._config[split])

            if spltype == 'dir':

                # deal with relative paths
                path = os.path.join(project_dir, splval) if not os.path.isabs(splval) else splval

                if not os.path.exists(path):

                    # raise if path does not exist
                    msg = f"Invalid value for option {split} - no such directory: {splval}"
                    suggestion = f"Please set {split} to a percentage, number or directory."

                    raise VergeMLError(msg, suggestion,
                                       hint_key=split, hint_type='value', help_topic='split')

    def _load_trained_model(self):
        """Load a trained models hyperparameters and results
        """

        train_mod_path = os.path.join(self._config['trainings-dir'], self.trained_model)
        if not os.path.exists(train_mod_path):
            raise VergeMLError("Trained model not found: {}".format(self.trained_model))

        # Merge data.yaml
        data_file = os.path.join(self._config['trainings-dir'], self.trained_model, 'data.yaml')
        if not os.path.exists(data_file):
            raise VergeMLError("data.yaml file not found for {}: {}".format(
                self.trained_model, data_file))

        doc = load_yaml_file(data_file, 'data file')
        self._config.update({
            'hyperparameters': doc.get('hyperparameters', {}),
            'results': doc.get('results', {}),
            'model': doc.get('model')
        })

        self.results = _Results(self, data_file)

        return data_file

    def _load_model_plugin(self, project_file, data_file, model_via_flag):
        if self._config['model']:
            # load the model plugin
            modelname = self._config['model']
            self.model_plugin = self.plugins.get("vergeml.model", modelname)

            if not self.model_plugin:
                message = f"Unknown model name '{modelname}'"
                suggestion = (did_you_mean(self.plugins.keys('vergeml.model'), modelname)
                              or "See 'ml help models'.")

                global ENV # pylint: disable=W0603

                # if model was passed in via --model
                if model_via_flag and ENV == self:
                    message = f"Invalid value for option --model: {message}"
                else:
                    res = None
                    if data_file:
                        # first check if model was defined in the data file
                        res = _check_definition(data_file, 'model', 'value')
                    if not res and project_file:
                        # next check the project file
                        res = _check_definition(project_file, 'model', 'value')
                    if res:
                        filename, definition = res
                        line, column, length = definition
                        # display a nice error message
                        message = display_err_in_file(filename, line, column,
                                                      f"{message} {suggestion}", length)
                        # set suggestion to None since it is now contained in message
                        suggestion = None

                raise VergeMLError(message, suggestion)
            else:
                # instantiate the model plugin
                self.model_plugin = self.model_plugin(modelname, self.plugins)

    def _load_yaml_and_configure(self, path, label, cache, device, device_memory): # pylint: disable=R0913
        doc = load_yaml_file(path, label)
        try:
            doc['device'] = parse_device(doc.get('device', {}),
                                         device_id=device,
                                         device_memory=device_memory)

            doc['data'] = parse_data(doc.get('data', {}), cache=cache, plugins=self.plugins)

            if 'random-seed' in doc and not isinstance(doc['random-seed'], int):
                raise VergeMLError('Invalid value option random-seed.',
                                   'random-seed must be an integer value.',
                                   hint_type='value',
                                   hint_key='random-seed')
        except VergeMLError as err:

            if err.hint_key:

                with open(path) as file:
                    definition = yaml_find_definition(file, err.hint_key, err.hint_type)

                if definition:
                    line, column, length = definition
                    err.message = display_err_in_file(path, line, column, str(err), length)
                    # clear suggestion because it is already contained in the error message.
                    err.suggestion = None
                    raise err
                else:
                    raise err
            else:
                raise err
        return doc

    def get(self, path):
        """Get a value by its path.

        For example, to access the variable 'id' in the dict 'device', use device.id as path.
        """
        return deepcopy(dict_get_path(self._config, path, None))

    def set(self, path, value):
        """Set a value by its path.
        """
        dict_set_path(self._config, path, value)

    def samples_dir(self):
        """Return the samples_dir or throw an error if it does not exist.
        """
        samples_dir = self._config['samples-dir']
        if not os.path.exists(samples_dir):
            raise VergeMLError(f'Could not find samples directory: {samples_dir}')
        elif not os.path.isdir(samples_dir):
            raise VergeMLError(f'Configured samples-dir is not a directory: {samples_dir}')
        return samples_dir

    def trained_model_dir(self):
        """Return the directory of a trained model (create if it does not exist).
        """
        assert self.trained_model, "You must create a trained model first."
        trained_model_dir = os.path.join(self._config['trainings-dir'], self.trained_model)
        try:
            os.makedirs(trained_model_dir)
        except FileExistsError:
            pass
        return trained_model_dir

    def checkpoints_dir(self):
        """Return the checkpoints directory (create if it does not exist)
        """
        checkpoints_dir = os.path.join(self.trained_model_dir(), "checkpoints")
        try:
            os.makedirs(checkpoints_dir)
        except FileExistsError:
            pass
        return checkpoints_dir

    def stats_dir(self):
        """Return the stats directory (create if it does not exist)
        """
        stats_dir = os.path.join(self.trained_model_dir(), "stats")
        try:
            os.makedirs(stats_dir)
        except FileExistsError:
            pass
        return stats_dir

    def cache_dir(self):
        """Return the cache directory (create if it does not exist)
        """
        cache_dir = self.get("cache-dir")
        try:
            os.makedirs(cache_dir)
        except FileExistsError:
            pass
        return cache_dir

    # save_max, min needed?
    def start_training(self,
                       name=None,
                       hyperparameters={}):
        """Start a training session.

        :param name: The name of the trained model to create
        :param hyperparameters: Pass in a models fixed parameters so they can be recovered later.
        """
        created = datetime.datetime.now()
        self.trained_model = name or random_robot_name(created, self._config['trainings-dir'])

        self.display.print("Creating @{} ...".format(self.trained_model))
        robot = ascii_robot(created, self.trained_model)
        if "VERGEML_FUNKY_ROBOTS" in os.environ:
            self.display.print("")
            self.display.print(robot)
            self.display.print("")
        else:
            self.display.print("")

        results = {'created-at': time.mktime(created.timetuple())}
        if self.get('data.input.type'):
            results['num-samples'] = self.data.num_samples('train')

        self.set('hyperparameters', hyperparameters)
        self.set('random-robot', robot)
        self.set('results', results)

        data_file = os.path.join(self.trained_model_dir(), "data.yaml")

        self.results = _Results(self, data_file)
        # results will do the rest. As results periodically updates data.yaml,
        # it will save hyperparameters too
        self.results.add({'status': "RUNNING", 'training-start': time.time()})
        self.results.flush()
        # create the training object
        stats_file = os.path.join(self.stats_dir(), "stats.csv")
        self.training = Training(self, stats_file)

        # save the original configuration
        config_file = os.path.join(self.trained_model_dir(), "vergeml.yaml")
        config = deepcopy(self._config)

        del config['hyperparameters']
        del config['results']
        del config['random-robot']

        if self.current_command:
            cmd, args = self.current_command
            config[cmd.name] = args

        with open(config_file, "w") as file:
            yaml.dump(config, file, default_flow_style=False)

        return self.trained_model

    def end_training(self, final_results={}):
        """End a training session.

        :param final_results: The final results of the training. This may include test accuracy, or the
                              final accuracy values if you save the best performing epoch.
        """
        assert self.training, "Must call start_training() first."
        final_results = deepcopy({k.replace('_', '-'):v for k, v in final_results.items()})
        self.results.add(final_results)
        self.results.add({'status': "FINISHED", 'training_end': time.time()})
        self.results.flush()
        self.training.update(write_stats=False, **final_results)
        self.training.end()
        self.training = None

    def cancel_training(self):
        """Cancel a training session.
        """
        assert self.training, "Must call start_training() first."
        self.results.add({'status': "CANCELED", 'training-end': time.time()})
        self.results.flush()
        self.training.end()
        self.training = None

    @property
    def data(self):
        """Return the data loader.
        """
        if not self._data:
            self._data = Data(self, plugins=self.plugins)
        return self._data

    def configure(self, *libraries):
        """Configure various libraries by setting up random seeds etc.

        :param libraries: The libraries to configure.
        """
        for library in libraries:
            assert library in ('keras', 'tensorflow', 'numpy', 'python', 'torch')
            if library == 'keras':
                KerasLibrary.setup(self)
            elif library == 'tensorflow':
                TensorFlowLibrary.setup(self)
            elif library == 'numpy':
                NumPyLibrary.setup(self)
            elif library == 'python':
                PythonInterpreter.setup(self)
            elif library == 'torch':
                TorchLibrary.setup(self)

    def progress_callback(self,
                          epochs=None,
                          steps=None,
                          display_progress='epochs-steps',
                          stats=_DEFAULT_STATS):
        """Get a generic callback to capture training progress.
        """
        assert self.training, "Must call start_training() before calling progress_callback()"
        assert display_progress in ('epochs', 'steps', 'epochs-steps', None)
        stats = deepcopy(stats)
        return self.training.callback(epochs, steps, display_progress, stats)

    def keras_callback(self,
                       display_progress='epochs-steps',
                       stats=_DEFAULT_STATS):
        """Get a callback suitable for passing to keras to get training feedback.
        """
        assert self.training, "Must call start_training() before calling keras_callback()"
        return KerasLibrary.callback(self, display_progress, stats)

    def tensorflow_session(self):
        """Create a new tensorflow session as configured in the environment.
        """
        return TensorFlowLibrary.create_session(self)

    def args_for(self, fn, args):
        """A utility function to return only the args which can be passed to function fn.

        Also adds base values from env, like random-seed and stats-dir
        """
        args = {k.replace("-", "_"):v for k, v in args.items()}
        res = {}
        params = inspect.signature(fn).parameters.keys()
        methods = {
            'stats_dir': self.stats_dir,
            'checkpoints_dir': self.checkpoints_dir,
            'samples_dir': self.samples_dir,
            'tensorflow_session': self.tensorflow_session,
            'trained_model_dir': self.trained_model_dir,
            'trained_model': lambda: self.trained_model,
            'test_split': lambda: self.get('test-split'),
            'val_split': lambda: self.get('val-split'),
            'cache_dir': self.cache_dir(),
            'random_seed': lambda: self.get('random-seed'),
            'trainings_dir': lambda: self.get('trainings-dir'),
        }
        for param in params:
            if param in args:
                res[param] = args[param]
            elif param in methods:
                res[param] = methods[param]()
        return res

    def set_defaults(self, cmd, args):
        """Set up environment defaults before executing the command.
        """
        if self.model_plugin:
            self.model_plugin.set_defaults(cmd, args, self)
            self._config['device'] = parse_device(self._config.get('device', {}))
            self._config['data'] = parse_data(self._config.get('data', {}))




def _check_definition(filename, key, kind):
    if not filename:
        return None
    with open(filename) as f:
        definition = yaml_find_definition(f, key, kind)
        return (definition, filename) if definition else None

_SYNC_INTV = 1.0 # sync interval in seconds
class _Results:
    def __init__(self, env, path):
        self.path = path
        self.env = env
        self.last_sync = None

    def add(self, data):
        for k, v in data.items():
            self.env.set('results.' + k.replace('_', '-'), v)
        self._sync()

    def flush(self):
        self._sync(force=True)

    def _sync(self, force=False):
        now = datetime.datetime.now()
        if force or not self.last_sync or (now - self.last_sync).total_seconds() > _SYNC_INTV:
            self.last_sync = now
            data = dict(model=self.env.get('model'),
                        hyperparameters=self._convert(self.env.get('hyperparameters')),
                        results=self._convert(self.env.get('results')))
            with open(self.path, "w") as f:
                yaml.dump(data, f, default_flow_style=False)

    def _convert(self, vals):
        res = {}
        for k, v in vals.items():
            if isinstance(v, (np.int, np.int8, np.int16, np.int32, np.int64)):
                v = int(v)
            elif isinstance(v, (np.float, np.float16, np.float32, np.float64)):
                v = float(v)
            elif isinstance(v, Labels):
                v = list(v)
            elif isinstance(v, np.ndarray):
                v = v.tolist()
            elif isinstance(v, dict):
                v = self._convert(v)
            res[k] = v
        return res

class Training:

    def __init__(self, env, stats_file):
        self.env = env
        self.steps = None
        self.epochs = None
        self.display_progress = None
        self.stats = None
        self.did_start = False
        self.training_feedback = None
        self.stats_file = stats_file
        self.stats_writer = None

    def callback(self, epochs, steps, display_progress, stats):
        self.steps = steps
        self.epochs = epochs
        self.display_progress = display_progress
        self.stats = stats
        self.current_step = 0
        self.current_epoch = 0
        self._avg = {}
        for stat in self.stats:
            if not 'title' in stat:
                stat['title'] = stat['name']
            if not 'category' in stat:
                stat['category'] = ''
            if not 'format' in stat:
                stat['format'] = "{.4f}"
            if not 'smooth' in stat:
                stat['smooth'] = False
            if not 'feedback' in stat:
                stat['feedback'] = True
            if not 'log' in stat:
                stat['log'] = True
        self.stats_writer = _StatsWriter(self.stats, self.stats_file)
        return self.update

    def update(self, epoch=None, step=None, write_stats=True, **stats):
        stats = deepcopy(stats)
        if epoch is not None:
            self.current_epoch = epoch
        if step is not None:
            self.current_step = step

        results = dict(epochs=self.current_epoch, steps=self.current_step)
        results.update(stats)
        self.env.results.add(results)

        if not self.did_start:
            if self.display_progress is not None:
                self.training_feedback = TrainingFeedback(epochs=self.epochs,
                                                          steps=self.steps,
                                                          display_progress=self.display_progress,
                                                          stats=self.stats,
                                                          display=self.env.display)
                self.training_feedback.start()
            self.did_start = True

        smooth_stats = deepcopy(stats)
        if stats:
            smooth = [stat['name'] for stat in self.stats if stat['smooth']]

            if step is not None:
                for k, v in stats.items():

                    if k in smooth:
                        self._avg.setdefault(k, [])
                        self._avg[k].append(v)
                        if len(self._avg[k]) > 10:
                            self._avg[k] = self._avg[k][-10:]
                        v = sum(self._avg[k]) / len(self._avg[k])

                    smooth_stats[k] = v
            if write_stats:
                self.stats_writer.write(epoch, step, stats)
            self.training_feedback.update(epoch=epoch, step=step, **smooth_stats)


    def end(self):
        if self.did_start:
            self.training_feedback.stop()
            self.stats_writer.end()


def _toscalar(v):
    if isinstance(v, (np.float16, np.float32, np.float64,
                      np.uint8, np.uint16, np.uint32, np.uint64,
                      np.int8, np.int16, np.int32, np.int64)):
        return np.asscalar(v)
    else:
        return v

class _StatsWriter:

    def __init__(self, stats, stats_file):
        self.ks = [k['name'] for k in stats if k['log']]
        self.prev = {}

        if not self.ks:
            return

        self.file = open(stats_file, "w", newline='')
        self.writer = csv.writer(self.file)
        self.writer.writerow(["epoch", "step"] + self.ks)

    def write(self, epoch, step, data):
        if not self.ks:
            return

        # Make sure that keys have no underscores.
        data = {k.replace('_', '-'):v for k, v in data.items()}

        row = [epoch, step]
        for k in self.ks:
            if k in data:
                row.append(_toscalar(data[k]))
                self.prev[k] = data[k]
            elif k in self.prev:
                row.append(_toscalar(self.prev[k]))
            else:
                row.append(None)

        self.writer.writerow(row)

    def end(self):
        if not self.ks:
            return

        self.file.flush()
        self.file.close()
        self.writer = None


from vergeml.operation import operation, OperationPlugin
from vergeml.option import option
from vergeml.utils import VergeMLError, SPLITS
from copy import copy

@operation('augment', descr="Augment a sample by producing multiple variants.", topic="general")
@option('variants', validate='>0', type=int)
class AugmentOperation(OperationPlugin):

    def __init__(self, variants, apply=None):
        super().__init__(apply)

        if not isinstance(variants, int):
            raise VergeMLError("The parameter 'variants' of 'augment' must be of type 'int'.")

        self.variants = variants

    def transform_sample(self, sample):
        if self.apply.intersection(set(SPLITS)) \
                and sample.meta['split'] not in self.apply:
            yield sample
        else:
            for _ in range(self.variants):
                yield copy(sample)

    def multiplier(self):
        return self.variants



from vergeml.plugins import PLUGINS
from vergeml.option import Option

_MODEL_META_KEY = '__vergeml_model__'

def model(name, descr=None, long_descr=None):
    """Define a command for your model.

    :param name:        Name of the operation.
    :param descr:       A short description of the operation
    :param long_descr:  A long description
    """
    def decorator(o):
        assert getattr(o, _MODEL_META_KEY, None) is None
        
        options = Option.discover(o)
        cmd = Model(name, 
                    descr=descr, 
                    long_descr=long_descr,
                    options=options)
        setattr(o, _MODEL_META_KEY, cmd)
        return o
    return decorator

class Model:
    def __init__(self, name, descr=None, long_descr=None, options=[], plugins=PLUGINS):
        self.name = name
        self.descr = descr
        self.long_descr = long_descr
        self.options = options
        self.plugins = plugins

    @staticmethod
    def discover(o, plugins=PLUGINS):
        res = None
        if hasattr(o, _MODEL_META_KEY):
            res = getattr(o, _MODEL_META_KEY)
            res.plugins = plugins
            for option in res.options:
                option.plugins = plugins
        return res


class ModelPlugin:

    def __init__(self, name, plugins=PLUGINS):
        self.name = name
        self.model = None
        self.plugins = plugins
    
    def load(self, env):
        raise NotImplementedError
    
    def project_file_template(self):
        return f"model: {self.name}" 
        
    def set_defaults(self, cmd, args, env):
        pass


from vergeml.utils import download_files
from vergeml.utils import Error
import shutil
import os.path
import zipfile
import tarfile


def download(env):
    """The Street View House Numbers (SVHN) Dataset.

SVHN is a real-world image dataset for developing machine 
learning and object recognition algorithms with minimal 
requirement on data preprocessing and formatting. It can be 
seen as similar in flavor to MNIST (e.g., the images are of 
small cropped digits), but incorporates an order of magnitude 
more labeled data (over 600,000 digit images) and comes from a 
significantly harder, unsolved, real world problem 
(recognizing digits and numbers in natural scene images). SVHN
 is obtained from house numbers in Google Street View images.



Authors:
  Yuval Netzer, Tao Wang, Adam Coates, Alessandro Bissacco, 
  Bo Wu, Andrew Y. Ng 
  Reading Digits in Natural Images with Unsupervised Feature 
  Learning NIPS Workshop on Deep Learning and Unsupervised 
  Feature Learning 2011. 

  http://ufldl.stanford.edu/housenumbers/nips2011_housenumbers.pdf

For more information visit: http://ufldl.stanford.edu/housenumbers/"""

    urls = ["http://ufldl.stanford.edu/housenumbers/train_32x32.mat",
            "http://ufldl.stanford.edu/housenumbers/test_32x32.mat"]

    samples_dir = env.get('base.samples_dir')
    print("Downloading SVHN to {}.".format(samples_dir))

    src_dir = download_files(urls, dir=env.get('base.cache_dir'))

    for file in ("train_32x32.mat", "test_32x32.mat", ):
        shutil.copy(os.path.join(src_dir, file), samples_dir)

    shutil.rmtree(src_dir)

    print("Finished downloading SVHN.")


download.__info__ = [
    ('Samples', '73K'),
    ('Test Samples', '26K'),
    ('Type', 'Labeled Images'),
    ('Resolution', '32x32 rgb'),
    ('Size', '235.3 MB')
]
